#ifndef BALANCER_H
#define BALANCER_H

#include <time.h>

/* - parameters for load-balancer --------------------------------------------------- */

#define ENTRY_EXPIRE_TIME	24	/* time for an entry to expire in hours */

/* ---------------------------------------------------------------------------------- */

#define HASH_TABLE_SIZE		1024	/* the hash table will have 1024 entries ... */

#define TASK_NAME_LENGTH	128	/* the maximum size of the task name + 1 */

#define BAL_QUEUE_EXPIRE	60	/* mark the task as dead after we do not found any other with this tid ... */

/*
one data structure holding statistical information about previous runs of a task
*/
struct bal_task_entry{
    char task_name[TASK_NAME_LENGTH];	/* NULL terminated name of the PVM process */
    time_t lastupdate;			/* timestamp of last update */
    
    double cpu_pct_avg;			/* average CPU percent usage */
    
    double total_avg;			/* average running time total in seconds */
    
    double mem_avg;			/* average memory usage in kilobytes */
    
    unsigned long of_writes;		/* number of recorded data states */
        
    struct bal_task_entry *next;	/* pointer to the next entry in a hash-table cell */
};

extern struct bal_task_entry *bal_table[HASH_TABLE_SIZE];

unsigned long bal_hash(unsigned char *str);
struct bal_task_entry *bal_search(char *name);
struct bal_task_entry *bal_insert(char *name);
void bal_delete(char *name);
void bal_table_init();
void bal_update_data(char *name,double pct_cpu,double lifespan,double mem_res);



/*
  data structure for holding the dead tasks queue, it is used to detect migrations of tasks,
  if a task is in this queue long enough, it is considered dead and moved to the HASH table
*/
struct bal_dead_queue_entry{
    char task_name[TASK_NAME_LENGTH];	/* name os the task*/
    int tid;				/* the tid - with help of this we can find the corresponding new task ... */
    time_t expire;			/* time of expiration - usually set to 30 seconds after termination */
    
    double pct_cpu;			/* CPU percent usage */
    
    double lifespan;			/* running time total in seconds */
    
    double mem_res;			/* memory usage in kilobytes */

    struct bal_dead_queue_entry *prev;	/* pointer to the previous entry ... */
};

void bal_queue_insert(char *name,int tid,double pct_cpu,double lifespan,double mem_res);
void bal_queue_purge();
int bal_queue_search(int tid);

#include "mon_util.h"

int bal_queue_search(int tid,struct procs_data *data);

#endif /* BALANCER_H */
