#include <stdio.h>         /* for fprintf */
#include <string.h>        /* for strerror */
#include <unistd.h>        /* for sysconf */
#include <stdlib.h>        /* for EXIT_FAILURE */
#include <assert.h>
#include <sys/types.h>     /* for processor_info */
#include <sys/sysinfo.h>   /* for cpu_sysinfo_t */

#include <sys/socket.h>
#include <netinet/in.h>    /* for in_addr */
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/time.h>      /* for gettimeofday */
#include <limits.h>        /* for PATH_MAX */


/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux
#include <kstat.h>         /* for kstat */
#include <sys/processor.h>
#include <sys/swap.h>      /* for swaptbl_t ... */
#include <sys/fault.h>
#else
#include <asm/page.h>

#define NR_TASKS	2048
//#include <linux/tasks.h> /* for NR_TASKS */


#define bytetok(x)	(((x) + 512) >> 10)
#define pagetok(x)	((x) << (PAGE_SHIFT - 10))
#include <errno.h>

#endif



#include <sys/signal.h>
#include <sys/syscall.h>
#include <sys/procfs.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <pthread.h>

#include <strings.h>

#include <algorithm>

#include "mon_slave.h"
#include "mon_util.h"
#include "mon_global.h"
#include "monitor.h"
#include "pvm3.h"



int tid;                      /* Global PVM task id */
int dtid;                     /* Task id of the local PVMD */

t_global_options global_options;

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
long mem_of_host; /* in kB */

void mon_sendmessage__ (int sockfd, message_id *mid, void *buffer, int bytes2write)
{
  int n;                    /* Bytes written by socket */

  XDR xdres_id;             /* XDR encode stream for message ident. */
  char *xdrbuff_id;         /* XDR buffer for message identification */
  int xdrbs_id;             /* XDR buffer size for message ident. */
  
   
  /* Allocate XDR buffer for the message identification */
  xdrbs_id = sizeof (message_id);
  if ((xdrbuff_id = 
       (char *) calloc (xdrbs_id, sizeof (char))) == NULL) {
    bailout ("mon_sendmessage(), calloc() for xdrbuff_id failed");
  }
  
  /* Create XDR stream for message identification */
  xdrmem_create (&xdres_id, xdrbuff_id, xdrbs_id, XDR_ENCODE);
  
  /* Encode the message type identifier */
  if (xdr_mtype_id (&xdres_id, mid) != TRUE) {
    bailout ("mon_sendmessage(), xdr_mtype_id() failed");
  }
  xdr_destroy(&xdres_id);
  
  /* Send message identification */
#ifndef OSTYPE_linux
  assert( *(int*)xdrbuff_id == PROCM );
#endif
  if ((n = write (sockfd, xdrbuff_id, xdrbs_id)) < 0) {
    bailout ("mon_sendmessage(), write() for xdrbuff_id failed");
  } else {
    if (n != xdrbs_id) {
      fprintf (stderr, "Wrote only %d bytes of %d bytes\n", n, xdrbs_id);
      bailout ("mon_sendmessage(), write() for xdrbuff_id failed");
    }
  }
  
  /* Send the buffer */
  if ((n = write (sockfd, buffer, bytes2write)) < 0) {
    bailout ("mon_sendmessage(), write() for buffer failed");
  } else {
    if (n != bytes2write) {
      fprintf (stderr, "Wrote only %d bytes of %d bytes\n", n, bytes2write);
      bailout ("mon_sendmessage(), write() for buffer failed");
    }
  }
#ifdef _DEBUG
  //  cerr<<getpid()<<"_bytes written: "<<n<<" of "<<bytes2write<<endl;
  cerr<<"mon_slave with pid: "<<getpid()<<" send a message\n";
#endif
}

/* Swap can be altered during uptime of the system */
void get_swap_info (nodeinfo *ni) 
{

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux
 
  int i;
  int sru;                   /* Swap resources in use */
  swaptbl_t *swpt_ptr;       /* Pointer to swap table */
  long tot_pg  = 0;          /* Total pages */
  long tot_fpg = 0;          /* Total free pages */
  int psize = getpagesize(); /* Number of bytes per page */

  /* Determine swap resources in use */
  if ((sru = swapctl (SC_GETNSWP, NULL)) == -1 ) {
    bailout ("get_installed_swap(), swapctl(SC_GETNSWP) failed");
  } else {
    
    /* Any swap space configured */
    if (sru > 0) {

      /* Allocate memory for the swaptable structure */
      if ((swpt_ptr = (swaptbl_t *) calloc (sizeof (swaptbl_t), 
					    sizeof (char))) == NULL) {
	bailout ("get_installed_swap(), calloc() for swpt_ptr failed");
      }
      
      /* Allocate memory for the dynamic members of the swaptable structure */
      for (i = 0; i < sru; i++) {
	if ((swpt_ptr->swt_ent[i].ste_path = 
	     (char *) calloc (PATH_MAX, sizeof (char))) == NULL) {
	  bailout ("get_installed_swap(), calloc() for ste_path failed");
	}
      }

      /* Get all swap entries */
      swpt_ptr->swt_n = sru;
      if ((sru = swapctl (SC_LIST, swpt_ptr)) < 0) {
	bailout ("get_installed_swap(), swapctl(SC_LIST) failed");
      }
      
      /* Calculate total swap space */
      for (i = 0; i < sru; i++) {
	tot_pg  += swpt_ptr->swt_ent[i].ste_pages;
	tot_fpg += swpt_ptr->swt_ent[i].ste_free;
      }
      
      /* Fill in the values for the node information */
      ni->tswap = (long) (tot_pg * psize / 1024);
      ni->fswap = (long) (tot_fpg * psize / 1024);
      ni->uswap = (long) ((tot_pg - tot_fpg) * psize / 1024);
      
      /* Free dynamically allocated memory */
      for (i = 0; i < sru; i++) {
	free ((void *) swpt_ptr->swt_ent[i].ste_path);
      }
      free ((void *) swpt_ptr);  

    } else {
      
      /* No swap installed */
      /* Fill in the values for the node information */
      ni->tswap = 0;
      ni->fswap = 0;
      ni->uswap = 0;
    }
  }
#else
/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
  char buffer[4096+1];
  int fd, len;
  char *p;

  fd = open("/proc/meminfo", O_RDONLY); /* ascii-file */
  len = read(fd, buffer, sizeof(buffer)-1);
  close(fd);
  buffer[len] = '\0';

  p = strchr(buffer, '\n'); /* get ride of the first lines */
  p = skip_token(p); /* "Mem:" */
  mem_of_host = strtol(p,&p,10)/1024;
  
  p = strchr(buffer, '\n'); /* forget the rest of the line */

  p = skip_token(p); /* "Swap:" */
  ni->tswap = strtoul(p,&p,10)/1024;
  ni->uswap = strtoul(p,&p,10)/1024;
  ni->fswap = strtoul(p,&p,10)/1024;
#endif
}
 

int get_static_data (nodeinfo *ni)
{
  int i;

  /* Determine number of CPUs */
  if ((ni->ncpu = sysconf(_SC_NPROCESSORS_CONF)) < 0) { 
    bailout ("get_static_data(), sysconf(_SC_NPROCESSORS_CONF) failed");
  }
  else {
    /* Allocate memory to store CPU loads */
    ni->bpcl = (cpuload **) malloc (sizeof (cpuload *) * ni->ncpu);
    for (i = 0; i < ni->ncpu; i++) {
      if ((ni->bpcl[i] = (cpuload *) calloc (1, sizeof(cpuload))) == NULL) {
	bailout ("get_static_data(), calloc() for cpuload failed");
      }
    }

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux
    /* Allocate memory for ks_name of CPUs */
    ni->bpkscn = (char **) malloc (sizeof (char *) * ni->ncpu);
    for (i = 0; i < ni->ncpu; i++) {
      if ((ni->bpkscn[i] = (char *) calloc (1, KSTAT_STRLEN)) == NULL) {
	bailout ("get_static_data(), calloc() for ks_name failed");
      }
    }
    
    /* Trace through the linked list to find the cpu names */
    if ((ni->kctrl = kstat_open()) != NULL) {
      for (i = 0, ni->klp = ni->kctrl->kc_chain; 
	   (i < ni->ncpu) && (ni->klp != NULL); 
	   ni->klp = ni->klp->ks_next) {
	if (strncmp(ni->klp->ks_name, "cpu_stat", 8) == 0) {
	  strcpy (ni->bpkscn[i++], ni->klp->ks_name);
	}
      }
      kstat_close(ni->kctrl);
    } else {
      bailout ("get_static_data(), kstat_open() failed");
    }
#endif
/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
  }

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux
  /* Get CPU type */
  if (processor_info(0, &(ni->pi)) < 0) {
    bailout ("get_static_data(), processor_info() failed");
  }
  else {  
#endif
/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
    /* Total phys. RAM */
    if ((ni->psize = sysconf(_SC_PAGESIZE)) < 0) {
      bailout ("get_static_data(), sysconf(_SC_PAGESIZE) failed");
    } else { 
      if ((ni->ppages = sysconf(_SC_PHYS_PAGES)) < 0) {
	bailout ("get_static_data(), sysconf(_SC_PHYS_PAGES) failed");
      } else { 
	ni->tram = (long) ((ni->psize * ni->ppages) / 1024);
      }
    }
/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux
  }
#endif
}


double
GetLoadAvrg( const char *average )
{
    assert( strcmp( average, "avenrun_1min" ) == 0  ||
            strcmp( average, "avenrun_5min" ) == 0  ||
            strcmp( average, "avenrun_15min" ) == 0 );

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
#ifndef OSTYPE_linux

    kstat_ctl_t *kc;
    kstat_t *ksp;
    kstat_named_t *kn;
 
    kc = kstat_open();
    if (kc == 0) {
        perror( "kstat_open failed" );
        return -1;
    }
 
    ksp = kstat_lookup(kc, "unix", 0, "system_misc");
    if (ksp == 0) {
        perror( "kstat_lookup failed" );
        kstat_close( kc );
        return -1;
    }

    if (kstat_read(kc, ksp,0) == -1) {
        perror( "kstat_read failed" );
        kstat_close( kc );
        return -1;
    }
 
    /* choose one of avenrun_1min, avenrun_5min, avenrun_15min */
    kn = (kstat_named_t*)kstat_data_lookup(ksp, average);
    if (kn == 0) {
        perror( "kstat_data_lookup failed" );
        kstat_close( kc );
        return -1;
    }

    kstat_close( kc );
    return (double)kn->value.ul/FSCALE;
#else
/* for LinuX, added by streit@upb.de at 22th Nov 1999 */

    char buffer[4096+1];
    int fd, len;
    char *p;
    fd = open("/proc/loadavg", O_RDONLY);
    len = read(fd, buffer, sizeof(buffer)-1);
    close(fd);
    buffer[len] = '\0';
    if (strcmp(average, "avenrun_15min" ) == 0) {
      p = skip_token(p); /* skip 1 min */
      p = skip_token(p); /* skip 5 min */
      return strtod(buffer, &p);
    } else if (strcmp(average, "avenrun_5min" ) == 0) {
      p = skip_token(p); /* skip 1 min */
      return strtod(buffer, &p);
    } else {
      /* in any case return the 1 min load average */
      return strtod(buffer, &p);
    }
    
#endif
}



int get_dynamic_data(nodeinfo *ni)
{
  static int runs = 2;   /* Flag for first execution, two times */

  double d;
  double e;
  int i;
  int j;
  

  /* Time stamp */
  if (gettimeofday(&(ni->ts), NULL) < 0) {
    bailout ("get_dynamic_data(), gettimeofday() failed");
  }
    
  /* Execute this twice at startup */
  do {

    double LoadAvrg = GetLoadAvrg( "avenrun_1min" );
    /* Calculate average CPU load */
    for (i = 0; i < ni->ncpu; i++) {
      ni->bpcl[i]->cl = LoadAvrg / ni->ncpu;
    }

    if (runs != 0) {
      sleep (1);
      runs--;
    }
    
  } while (runs);


  /* RAM used */
  if (ni->ppages <= 0) { 
    bailout ("get_dynamic_data(), No physical RAM, no used RAM");
  } else { 
    if ((ni->apages = sysconf(_SC_AVPHYS_PAGES)) < 0) { 
      bailout ("get_dynamic_data(), no physical RAM avaialble\n");
    } else {
      ni->fram = ni->apages * ni->psize / 1024;
      ni->uram = (ni->ppages - ni->apages) * ni->psize / 1024;
    }
  }
  
  /* Swap information */
  get_swap_info (ni);
}



int connect_to_master() 
{
  int i;

  struct hostent *hp, he;       /* Host entry */
  int herr;                     /* Error status from gethostbyname_r */
  char hbuff[MAXHOSTNAMELEN+1]; /* Hostname */
  int family;                   /* Protocol family */
  struct in_addr saddr;         /* Protocol specific data structure */
  int rsa = 1;                  /* Reuse local socket address */
  int sfd = -1;                 /* Socket file descriptor */
  struct sockaddr_in sin;       /* Protocol specific data structure */
  int err;                      /* Error from connect */

  /* Get address by name */
#ifndef OSTYPE_linux
  if ((hp = gethostbyname_r(global_options.MonHost, 
			    &he, 
			    hbuff, 
			    sizeof(hbuff), 
			    &herr)) == NULL) {
#else
    /* gethostbyname_r doesnt exist in Linux */
  if ((hp = gethostbyname(global_options.MonHost)) == NULL) {
#endif
    bailout ("connect_to_master(), gethostbyname_r(): failed");
  } else {
#ifdef _DEBUG
    cerr<<"connect_to_master: hostname of master: "<<hp->h_name<<endl;
#endif
    family = hp->h_addrtype;
    bcopy(hp->h_addr, (char *) &saddr, hp->h_length);

    /* Create local socket */
    if ((sfd = socket(family, SOCK_STREAM, 0)) < 0) {
      bailout ("connect_to_master(), socket() failed");
    }
    /* shouldnt be here in LINUX, maybe in SOLARIS -> test it */
    //sfd = htons(sfd);
    
    /* Reuse socket */
    if (setsockopt(sfd, 
		   SOL_SOCKET,
		   SO_REUSEADDR, 
		   (char *) &rsa, 
		   sizeof (rsa)) < 0) {
      bailout ("connect_to_master(), setsockopt() failed");
    }

    sin.sin_addr = saddr;
    sin.sin_port = htons(global_options.MonPort);
#ifdef HAS_SINLEN
    sin.sin_len = sizeof (struct sockaddr_in);
#endif
    sin.sin_family = family;

    /* Try to connect to the master */
    for (i = 0; i < CONRETRCNT; i++) {
#ifdef _DEBUG
      cerr<<"number of attempts: "<<i<<endl;
      err = connect(sfd, (struct sockaddr *) &sin, sizeof(sin));
      perror("connect_to_master");
      if (err < 0) {
#else
      if ((err = connect(sfd, (struct sockaddr *) &sin, sizeof(sin))) < 0) {
#endif
	/* Error connecting to master, wait a little time */
	sleep (CONRETRDEL);
      } else {
	break;
      }
    }
    /* Are we connected to the master */
    if (err < 0) {
      close(sfd);
      bailout ("connect_to_master(), connect() failed");
    }
  }
  return sfd;
}
 
void dump_node_data (lm_data *mb)
{
#ifdef DO_DUMPS
  static int firstrun   = 1;

  FILE *f;
  int size;
  char *name;

  /* Create unique name */  
  size = strlen (global_options.DumpDir) + strlen (mb->hname) + 2;
  if ((name = (char *) calloc (size, sizeof (char))) == NULL) {
    bailout ("dump_node_data(), calloc() failed");
  } else {
    if (sprintf (name, "%s/%s", global_options.DumpDir, mb->hname) < 1) {
      bailout ("dump_node_data(), sprintf() failed");
    }
  }
  /* Open file for writing */
  if ((f = fopen (name, "a")) == NULL) {
    bailout ("dump_node_data(), open() failed");
  }
  free( name );

  /* Dump data to file */
  while (firstrun) {
    fprintf (f, "Host   HID  #CPU   load     kernel     wait      user      idle    uram   fram   uswap  fswap    ts\n");
    firstrun--;
  }
  fprintf (f, 
	   "%.5s %3d   %3d %9.6f %9.6f %9.6f %9.6f %9.6f %6d %6d %6d %6d %6d\n",
	   mb->hname,
	   mb->hid,
	   mb->ncpu,
	   mb->cpuload,
	   0,  /* mb->kernel, */
	   0,  /* mb->wait, */
	   0,  /* mb->user, */
	   0,  /* mb->idle, */
	   mb->uram,
	   mb->fram,
	   mb->uswap,
	   mb->fswap,
	   mb->ts
	   );

  /* Close the file */
  fclose (f);
#endif
}

volatile int timer_expired;


void mon_handle_alarm(int signum){
  if(signum==SIGALRM){
    timer_expired=1;
  }
}



int main (int argc, char *argv[])
{
  int i;
  int j;

  double d;
  double e;

  nodeinfo ni;

  lm_data mb;                   /* Message buffer for data */

  message_id  mb_id;            /* Message buffer for message ident. */

  int sfd;                      /* Socket file descriptor */

  XDR xdres;                    /* XDR encode stream for load data */
  char *xdrbuff;                /* XDR buffer for load data */
  int xdrbs;                    /* XDR buffer size for load data */

  XDR xdres_id;                 /* XDR encode stream for message ident. */
  char *xdrbuff_id;             /* XDR buffer for message identification */
  int xdrbs_id;                 /* XDR buffer size for message ident. */

  int n;                        /* Bytes written */

  int mon_interval;             /* Monitoring interval */

  int info;                     /* PVM error flag */

  int ntask;                    /* Number of tasks */
  struct pvmtaskinfo *taskaddr; /* Base pointer to pvmtaskinfo structure */

  double ibase2[16];            /* Inverse of the power to 2 */
  char path[PATH_MAX];          /* Path to the /proc file */
  int f;                        /* File descriptor of /proc */
  prpsinfo_t psinfo;            /* Process info structure */

  procs_data procsdata;         /* Process information */

  struct timeval ts;            /* Time stamp */

  message_id mid;               /* The message identification */

  message_id mid2;               /* The second message identification for the TERMINATED PROCESS messages */

  pid_t pid;                    /* Local process id of process monitor */

  char *host;   /* The hostname of the master monitor */
  int port;     /* The port of the master monitor */

  int procsmon_enabled=0;

  resdes *rd_base = NULL;   	/* Base pointer to resource description */

  double last_cpuload=-1;

#ifdef USE_HASH_DATA
  int watch_array[512];		/* array of tids to watch for termination ... */
  int watch_array_old[512];	/* backup array of watch_array */
  int of_watch_tasks=0;		/* how many tids are in the watch array ? */
  int of_watch_tasks_old=0;	/* backup counter of of_watch_tasks*/
  
  struct timespec req;		/* structure for nanosleep() */
  int bufid;			/* id for the PVM receive buffer */
  int died_tid;			/* variable holding the tid of a just exited process */
#endif
  
  int hid = -1;                   /* Host id from resource file */
  char *hname;                    /* Local hostname */
  
  char proc_name[DEFSTR];

#ifdef OSTYPE_linux
  static struct save_hist save_history[NR_TASKS];
  struct save_hist New_save_history[NR_TASKS];
  int max_procs = 0;
  int count_now = 0;
#endif

#ifdef _DEBUG
  cerr<<getpid()<<"_debug: slave with pid "<<getpid()<<endl;
  cerr<<getpid()<<"_debug: argc "<<argc<<endl
      <<getpid()<<"_debug: argv[0] "<<argv[0]<<endl
      <<getpid()<<"_debug: argv[1] "<<argv[1]<<endl
      <<getpid()<<"_debug: argv[2] "<<argv[2]<<endl
      <<getpid()<<"_debug: argv[3] "<<argv[3]<<endl
      <<getpid()<<"_debug: argv[4] "<<argv[4]<<endl;

#endif

  //sleep(30);

  /* Check command line syntax */
  if( argc != 4 ) {
	bailout( "usage: mon_slave master_host master_port config_file\n" );
  }

  global_options.MonPort = atoi( argv[2] );
  if( global_options.MonPort <= 0 ) {
      bailout( "Invalid master monitor port" );
  }
  strcpy( global_options.MonHost, argv[1] );

  /* Alocate memory for hostname */
  if ((hname = (char *) calloc (MAXHOSTNAMELEN + 1, sizeof (char))) == NULL) {
    bailout ("mon_procs(), calloc() for hname failed");
  }
  /* Store the hostname */
  if (gethostname(hname, MAXHOSTNAMELEN + 1) < 0) {
    bailout ("mon_procs(), gethostname() failed");
  }

  mon_interval = DEFMONINT;

  /* Read options from config file */
  read_options( &global_options, argv[3] );

  /* Read resource description */
  if (read_resdes (&rd_base, argv[3]) < 1) {
    bailout ("main(), can't parse resource description");
  }

  /* Collect the static data of the node */
  get_static_data(&ni);
#ifdef _DEBUG
  cerr<<getpid()<<"-- "
      <<"#cpu:"<<ni.ncpu<<" "
      <<"psize:"<<ni.psize<<" "
      <<"pages:"<<ni.ppages<<" "
//       <<"avail-pages:"<<ni.apages<<" "
      <<"tram:"<<ni.tram<<" ";
//       <<"uram:"<<ni.uram<<" "
//       <<"fram:"<<ni.fram<<" "
//       <<"tswap:"<<ni.tswap<<" "
//       <<"uswap:"<<ni.uswap<<" "
//       <<"fswap:"<<ni.fswap<<" ";
//   for (i = 0; i < ni.ncpu; i++) {
//     cerr<<"cl:"<<ni.bpcl[i]->cl<<" "    
// 	<<"idle:"<<ni.bpcl[i]->idle<<" "
// 	<<"wait:"<<ni.bpcl[i]->wait<<" "
// 	<<"user:"<<ni.bpcl[i]->user<<" "
// 	<<"kernel:"<<ni.bpcl[i]->kernel<<" ";
//   }
  cerr<<endl;
#endif
  /* Connect to monitor master */
  if ((sfd = connect_to_master()) < 0) {
    bailout ("main(), connect_to_master() failed");
  }
  
  /* Fill in the static data which has to be send */
  if ((mb.hname = (char *) malloc ((size_t) (MAXHOSTNAMELEN + 1))) == NULL) {
    bailout ("main(), malloc() for sname failed");
  }
  
  /* Store the hostname */
  if (gethostname(mb.hname, MAXHOSTNAMELEN + 1) < 0) {
    bailout ("main(), gethostname() failed");
  }

  /* Get host id from resource entries */
  if ((mb.hid = get_resdes_hid (rd_base, mb.hname)) < 0) {
    bailout ("get_resdes_hid() failed");
  }
  hid = mb.hid;
 
  /* Get local process id */
  if ((pid = getpid()) < 0) {
    bailout (",getpid() failed");
  }

  /* Allocate memory for XDR buffer */
  if ((xdrbuff = (char *) calloc (sizeof (procs_data) + MAXHOSTNAMELEN + 1, 
				  sizeof (char))) == NULL) {
    bailout ("mon_procs(), calloc() for xdrbuff failed");
  }
  
  /* Initialize the array with the inverse of the power to 2 */
  for (i = 1, ibase2[0] = 1; i < 16; ibase2[i] = ibase2[i-1] / 2.0, i++);

  /* Initialize message identification tag */
  mid.mtype = PROCM;

  /* Initialize second message identification tag */
  mid2.mtype = TERMS;

  /* Store the number of CPUs */
  mb.ncpu = ni.ncpu;
    
  /* Fill in the type of each message */
  mb_id.mtype = LOADM;
  
  /* Allocate XDR buffer for the message identification */
  xdrbs_id = sizeof (message_id);
  if ((xdrbuff_id = (char *) calloc (xdrbs_id, sizeof (char))) == NULL) {
    bailout ("main(), calloc() for xdrbuff_id failed");
  }

  /* Create XDR stream for message identification */
  xdrmem_create (&xdres_id, xdrbuff_id, xdrbs_id, XDR_ENCODE);
  /* Encode the messege type identifier */
  if (xdr_mtype_id (&xdres_id, &mb_id) != TRUE) {
    bailout ("main(), xdr_mtype_id() failed");
  }
  xdr_destroy(&xdres_id);
  
  /* Allocate XDR buffer for the data */
  xdrbs = sizeof (lm_data) + MAXHOSTNAMELEN + 1;
  if ((xdrbuff = (char *) calloc (xdrbs, sizeof (char))) == NULL) {
    bailout ("main(), calloc() for xdrbuff failed");
  }

  signal(SIGALRM,mon_handle_alarm);
  
  /* call the former procsmon functions */
 
  while (1) {
    if (pvm_is_running() && !procsmon_enabled) {
	/* Enroll this task into PVM */
	if ((tid = pvm_mytid()) < 0) {
		bailout ("mon_procs(), pvm_mytid() failed");
	}
  	/* Get local PVMD task id */
	if ((dtid = pvm_tidtohost(tid)) < 0) {
		bailout ("mon_procs(), pvm_tidtohost() failed");
	}
        signal(SIGURG, SIG_IGN);
  	procsmon_enabled=1;
    }

    do {
        /* Collect the dynamic data */

        get_dynamic_data(&ni);

        /* Prepare the dynamic data which has to be send */

        /* CPU statistics */
        mb.cpuload = 0;
        for (i = 0; i < ni.ncpu; i++) {
            mb.cpuload +=  ni.bpcl[i]->cl;
        }

        /* Check for NaN's (Not a Number) */
    } while (mb.cpuload != mb.cpuload);
    
    if(last_cpuload<0)last_cpuload=mb.cpuload;
    mb.trend=mb.cpuload-last_cpuload;
    last_cpuload=mb.cpuload;

#ifdef USE_HASH_DATA
    // so far, there are no tids in the watch array
    of_watch_tasks=0;
#endif

    /* Physical RAM */
    mb.uram = ni.uram;
    mb.fram = ni.fram;
    
    /* Swap space */
    mb.uswap = ni.uswap;
    mb.fswap = ni.fswap;

    /* Timestamp */
    mb.ts = ni.ts.tv_sec;      
 
    /* Create XDR stream for data message */
    xdrmem_create (&xdres, xdrbuff, xdrbs, XDR_ENCODE);
    
    /* Encode the data messege */
    if (xdr_lm_data (&xdres, &mb) != TRUE) {
      bailout ("main(), xdr_lm_data() failed");
    }
    xdr_destroy(&xdres);
    /* Send message identification */
#ifndef OSTYPE_linux
    assert( *(int*)xdrbuff_id == LOADM );
#endif
    if ((n = write(sfd, xdrbuff_id, xdrbs_id)) < 0) {
      bailout ("main(), write() for xdrbuff_id failed");
    }
    
    /* Send data message */
    if ((n = write(sfd, xdrbuff, xdrbs)) < 0) {
      bailout ("main(), write() for xdrbuff failed");
    } else {
      if (n != xdrbs) {
	fprintf (stderr, 
		 "main(), write() wrote only %d bytes of %d bytes\n", 
		 n,
		 xdrbs);
	exit (EXIT_FAILURE);
      }
    }
    /* Dump the current data to file */
    dump_node_data (&mb);

    /* call procsmon functions */
    if (procsmon_enabled) {
      /* Get information of local PVM tasks from PVMD */
      if ((info = pvm_tasks (dtid, &ntask, &taskaddr)) < 0)
	   bailout ("mon_procs(), pvm_tasks() failed");


#ifdef OSTYPE_linux
      /* get time now */
      struct timeval update_time;
      gettimeofday(&update_time,0);
      //      time_t update_time;
      //      time(&update_time);
      double time_now = (update_time.tv_sec + update_time.tv_usec * 1e-6);
      double static last_update_time = time_now + 1e-9;
      double elapsed_time = time_now - last_update_time;
#endif

      /* Get the process statistics of each task */
      for (i = 0; i < ntask; i++) {
        /* Initialize structure */
        bzero (&procsdata, sizeof (procs_data));
        /* Get the process id */
        procsdata.pid = taskaddr[i].ti_pid;
        /* The monitor itself is not movable */
#ifdef _DEBUG
	cerr<<"##itself## "<<procsdata.pid<<" != "<<pid<<" --> "
	    <<(procsdata.pid != pid )<<endl<<flush;
#endif
        procsdata.movable = (procsdata.pid != pid );
        /* Orphans must also not be moved */
        if( taskaddr[i].ti_ptid == 0 ) procsdata.movable = FALSE;
	/* pvmgs isn't moveable */
	char *taskname = taskaddr[i].ti_a_out;
	if (strcmp(taskname,"pvmgs") == 0) procsdata.movable = FALSE;
#ifdef _DEBUG
	if (strcmp(taskname,"pvmgs") == 0) {
	  cerr<<"tid: "<<hex<<taskaddr[i].ti_tid
	      <<", name: "<<taskname
	      <<", compare: "<<dec<<strcmp(taskname,"pvmgs")
	      <<", is moveable: "<<procsdata.movable<<endl<<flush;
	}
#endif
#ifdef USE_HASH_DATA
	/* if the task is moveable, we will include it in the watch_array */
	if(procsdata.movable){
	  watch_array[of_watch_tasks]=taskaddr[i].ti_tid;
	  of_watch_tasks++;
	}
#endif	
#ifdef _DEBUG
        if( !procsdata.movable &&  procsdata.pid!=getpid() ) {
	  cerr << "immobile task 1: " << hex<< taskaddr[i].ti_tid  
	       << " ti_ptid==" << hex << taskaddr[i].ti_ptid << endl;
        } else if( procsdata.movable ) {
	  cerr << "Task is movable " << hex<< taskaddr[i].ti_tid << endl;
        }
#endif

#ifdef _DEBUG
	cerr<<"##orphans## ptid: "<<taskaddr[i].ti_ptid
	    <<", pid: "<<taskaddr[i].ti_pid
	    <<", tid: "<<taskaddr[i].ti_tid
	    <<" --> movable: "<<procsdata.movable<<endl<<flush;
#endif
	
	/* Set host id from resource file */
	procsdata.hid = hid;
	/* Get task id */
	procsdata.tid = taskaddr[i].ti_tid;
	strcpy(procsdata.pname,taskaddr[i].ti_a_out); /* save the process name */
	
#ifndef OSTYPE_linux
	/* Where we will find our process statistics */
	if (sprintf (path, "/proc/%d", procsdata.pid) < 0) {
	  bailout ("mon_procs(), sprintf() failed");
        }
	
	/* Get statistics of process */
	if ((f = open (path, O_RDONLY)) < 0) {
	  /* If opening the file failed that has not to be an error, 
	     maybe the process terminated. Continue with the next one */
	  continue;
	} else {
	  if (ioctl(f ,PIOCPSINFO, &psinfo) < 0) {
	    bailout ("mon_procs(), ioctl() failed");
	  }
	  close (f);
	}
	for (j = 0, procsdata.pct_cpu = 0, procsdata.pct_mem = 0; 
	     j < 16; j++) {
	  
	  /* Calculate cpu statistics of process */
	  procsdata.pct_cpu += 
	    ((psinfo.pr_pctcpu >> (15 - j)) & 0x0001) * ibase2[j];
	  
	  /* Calculate memory statistics of process */
	  procsdata.pct_mem += 
	    ((psinfo.pr_pctmem >> (15 - j)) & 0x0001) * ibase2[j];
	}
	
	/* Get size of process image and used memory resources */
	procsdata.mem_img = psinfo.pr_bysize / 1024;
	procsdata.mem_res = psinfo.pr_byrssize / 1024;	
	
#else
	/* for LinuX, added by streit@upb.de at 23th Nov 1999 */
	
	char buffer[4096+1];
	int fd, len;
	char *p;
	
	/* get the uptime of the system */
	fd = open("/proc/uptime", O_RDONLY); /* ascii-file */
	if (fd == 0) {
	  bailout("mon_procs(), getting uptime of system failed");
	}
	len = read(fd, buffer, sizeof(buffer)-1);
	close(fd);
	buffer[len] = '\0';
	p = buffer;
	double GL_current_time = strtod(p,&p);
	
	/* all statistics about a process are in an ascii file */
	if (sprintf(path,"/proc/%d/stat",procsdata.pid) < 0) {
	  bailout("mon_procs(), sprintf() failed");
	}
	
	/* the file in /proc */
	fd = open(path, O_RDONLY);
	if (fd <= 0) {
	  /* maybe process terminated -> just continue */
#ifdef _DEBUG
	  cerr<<"CONTINUE\n"<<flush;
#endif
	  continue;
	} else {
	  
	  len = read(fd, buffer, sizeof(buffer)-1);
	  close(fd);
	  buffer[len] = '\0';
	  
	  /* now all information are in buffer */

	  char *q = strchr(buffer,'(');
	  p = q+1;
	  	  
	  p = read_procname(p,proc_name);
	  
	  	  
	  p = skip_ws(p); /* process name */
	  
	  p = skip_token(p); /* skip pstate */
	  p = skip_token(p); /* skip ppid */
	  p = skip_token(p); /* skip pgrp */
	  p = skip_token(p); /* skip session */
	  p = skip_token(p); /* skip tty */
	  p = skip_token(p); /* skip tty pgrp */
	  p = skip_token(p); /* skip flags */
	  p = skip_token(p); /* skip min flt */
	  p = skip_token(p); /* skip cmin flt */
	  p = skip_token(p); /* skip maj flt */
	  p = skip_token(p); /* skip cmaj flt */
	  long utime = strtol(p,&p,10); /* utime in msec */
	  long stime = strtol(p,&p,10); /* stime in msec */
	  p = skip_token(p); /* skip cutime */
	  p = skip_token(p); /* skip cstime */
	  p = skip_token(p); /* skip priority */
	  p = skip_token(p); /* skip nice */
	  p = skip_token(p); /* skip timeout */
	  p = skip_token(p); /* skip it_real_val */
	  long proc_starttime = strtol(p,&p,10); /* start time in sec from 1.1.70 */
	  long GL_time_now = time(0L);
	  double proc_life_time = ((HZ*GL_current_time)-proc_starttime)/HZ;
	  
	  //p = skip_token(p); /* skip priority */
	  //p = skip_token(p); /* skip nice */
	  //p = skip_token(p); /* skip timeout */
	  //p = skip_token(p); /* skip it_real_val */
	  //p = skip_token(p); /* skip start_time */
	  /* get image size and resident set size */
	  procsdata.mem_img = bytetok(strtoul(p,&p,10)); /* image size */
	  procsdata.mem_res = pagetok(strtoul(p,&p,10)); /* resident set size */
	  
	  /* percentage of mem usage */
	  procsdata.pct_mem = ((double)procsdata.mem_img / 
			       (double)mem_of_host);
			       
	  
#ifdef _DEBUG
	  cerr<<getpid()
	      <<"_pid: "<<procsdata.pid
	      <<" --> %MEM: "<<procsdata.pct_mem<<" %\n";
#endif
	  /********************/
	  /* compute the %cpu */
	  /********************/
	  int n,i,index = 0;
	  int process_id = procsdata.pid;
	  double pcpu = 0.0;
	  
	  /* find pid in save_history */
	  i = 0;
	  int found = -1;

	  while (i < max_procs) {
#ifdef _DEBUG
	    cerr<<getpid()<<" investigating #"<<i
		<<" with PID "<<save_history[i].pid
		<<" and looking for PID "<<process_id<<endl<<flush;
#endif
	    if (save_history[i].pid == process_id) {
	      found = i;
	      i = max_procs;
	    }
	    i++;
	  }
	  
	  if (found > -1) {
#ifdef _DEBUG
	    cerr<<getpid()<<"_process "<<process_id<<" not new, position: "
		<<found<<endl<<flush;
#endif
	    /* process is not new */
#ifdef DEBUG
	    if (process_id == 826) {
	      cerr<<"elapsed_time: "<<elapsed_time<<" of "<<PCPUINT<<endl<<flush;
	    }
#endif
	    if (elapsed_time >= PCPUINT) {
	      /* compute new value for pcpu */
	      // correct the elapsed time
	      last_update_time = time_now;
	      
#ifdef DEBUG
	      if (process_id == 826) {
		cerr<<"new: "<<utime<<", "<<stime<<"  -- old: "
		    <<save_history[found].utime<<", "<<save_history[found].stime<<endl<<flush;
	      }
#endif
	      // compute pcpu
	      pcpu = (((double)( (utime-save_history[found].utime)+
				 (stime-save_history[found].stime)))/HZ) / elapsed_time;
#ifdef DEBUG
	      if (process_id == 826) {
		cerr<<"--> "<<(((double)( (utime-save_history[found].utime)+
					  (stime-save_history[found].stime)))/HZ)
		    <<" s of "<<elapsed_time<<" s"<<endl<<flush;
	      }
#endif

	      // correct utime and stime
	      save_history[found].utime = utime;
	      save_history[found].stime = stime;
	      	      
	      // store values
	      //save_history[found].utime = utime;
	      //save_history[found].stime = stime;
	      //save_history[found].pcpu = pcpu;
	    } else {
	      /** too fast, no recomputation necessary **/
	      utime = save_history[found].utime;
	      stime = save_history[found].stime;
	      pcpu = save_history[found].pcpu;
	    }
	    procsdata.totals=elapsed_time;
	    
	  } else {
#ifdef _DEBUG
	    cerr<<getpid()<<"_process "<<process_id<<" is NEW"<<endl<<flush;
#endif
	    /* new process */
#ifdef DEBUG
	    if (process_id == 826) {
	      cerr<<"NEW -> "<<(((double)(utime+stime))/HZ)
		  <<" s of "<<proc_life_time<<" s"<<endl<<flush;
	    }
#endif
	    /* compute the value */
	    pcpu = (((double)(utime+stime))/HZ) / proc_life_time;
	    
	    procsdata.totals=proc_life_time;
	    // create a new entry in save_history
	    //save_history[prev_count].pid = process_id;
	    //save_history[prev_count].utime = utime;
	    //save_history[prev_count].stime = stime;
	    //save_history[prev_count].pcpu = pcpu;

	    // updating the counter
	    //prev_count++;
	  }
#ifdef _DEBUG
          cerr<<getpid()<<"_PID "<<process_id<<" ->"<<(((double)(utime+stime))/HZ)
              <<" s of "<<proc_life_time<<" s used ---> "
              <<pcpu<<endl<<flush;
#endif
	  /* :-) */
	  if (pcpu < 0) pcpu = 0;
	  //if (pcpu > 1) pcpu = 1;
	  procsdata.pct_cpu = pcpu;
	  
#ifdef _DEBUG
	  cerr<<getpid()<<"_CPU usage for process "<<process_id
	      <<" is "<<pcpu<<endl<<flush;
#endif

	  if (count_now > NR_TASKS) {
	    bailout("mon_procs(), too many tasks");
	  }
#ifdef _DEBUG
	  cerr<<"storing at position "<<count_now
	      <<" for PID "<<process_id<<" "
	      <<utime<<" "
	      <<stime<<" "
	      <<pcpu<<endl<<flush;
#endif
	  // store values
	  New_save_history[count_now].pid = process_id;
	  New_save_history[count_now].tid = procsdata.tid;
	  New_save_history[count_now].utime = utime;
	  New_save_history[count_now].stime = stime;
	  New_save_history[count_now].pcpu = pcpu;
	  New_save_history[count_now].mem_res = procsdata.mem_res;
	  count_now++;
	} // if (fd <= 0)
#endif // Solaris <-> Linux
#ifdef _DEBUG
	if( !procsdata.movable &&  procsdata.pid!=getpid() ) {
	  cerr << "immobile task 2: " << procsdata.tid << endl;
	}
#endif
	// Correct load value, if the host is overloaded
	double load = GetLoadAvrg( "avenrun_1min" );
	if( load > 1 ) {
	  procsdata.pct_cpu *= load;
	  //	  procsdata.pct_cpu = min( 1.0, procsdata.pct_cpu*load );
	}
	
	/* Copy hostname */ 
	if ((procsdata.hname = strdup (hname)) == NULL) {
	  bailout ("mon_procs(), strdup() for hname failed");
	}
	/* Get timestamp */
	if (gettimeofday(&ts, NULL) < 0) {
	  bailout ("mon_procs(), gettimeofday() failed");
	}
	/* We are only interested in seconds */
	procsdata.ts = ts.tv_sec;
	
#ifdef _DEBUG
	cerr<<getpid()<<"-- "
	    <<procsdata.hname<<" "
	    <<procsdata.hid<<" "
	    <<procsdata.tid<<" *"
	    <<procsdata.pid<<" *"
	    <<procsdata.pct_cpu<<" *"
	    <<procsdata.pct_mem<<" *"
	    <<procsdata.mem_img<<" *"
	    <<procsdata.mem_res<<" "
	    <<procsdata.ts<<" "
	    <<procsdata.movable<<" "<<endl;
#endif
	  
	/* Encode the data into the XDR stream */
#ifdef _DEBUG
	if( !procsdata.movable &&  procsdata.pid!=getpid() ) {
	  cerr << "immobile task 3: " << procsdata.tid << endl;
	}
#endif
	xdrmem_create (&xdres, xdrbuff, sizeof (procs_data) + MAXHOSTNAMELEN + 1, XDR_ENCODE);
	xdr_procs_data (&xdres, &procsdata);
	xdr_destroy (&xdres);
	/* Send statistics to load master */
	mon_sendmessage__ (sfd, &mid, xdrbuff, 
			   sizeof (procs_data) + MAXHOSTNAMELEN + 1);
	/* Free the allocated memory */
	free (procsdata.hname);
      }
    } /* procsmon */

#ifdef USE_HASH_DATA
/* now look for the processes that we waited for, but did not receive a PVM notify event of
   dying nor found them in the current virtual machine task enrollment */
    
    for(int j=0;j<of_watch_tasks_old;j++){
      if(watch_array_old[j]>=0){
	/* try to find the tid in the New_save_history*/
	int i=0;
	while(i<count_now){
	  /* if we find it, just skip the tests */
	  if(New_save_history[i].tid==watch_array_old[j])break;
	  i++;
	}
	
	/* wuf, we watched for it, but now it's gone ... let's notify about that, drawing data from the old save_history */
	if(i>=count_now){
	  int found=-1;
	  i=0;
	  while(i<max_procs){
	    if(save_history[i].tid==watch_array_old[j]){
	      found=i;
	      break;
	    }
	    i++;
	  }
	  /* we have found the old task entry, so just send it ... */
  	  if(found>=0){
	  
	    /* fill in the values ... for now, just the last known we had */
	    procsdata.hname=strdup(hname);
	    procsdata.pid=save_history[found].pid;
	    procsdata.tid=save_history[found].tid;
	  
	    procsdata.pct_cpu=save_history[found].pcpu;
	    procsdata.pct_mem=0;
	    procsdata.mem_res=save_history[found].mem_res;
	  
	    /* always 1, since we monitor only moveable tasks */
	    procsdata.movable=1;
	  
	    /* Get timestamp */
	    if (gettimeofday(&ts, NULL) < 0) {
	      bailout ("mon_procs(), gettimeofday() failed");
	    }
	    /* We are only interested in seconds */
	    procsdata.ts = ts.tv_sec;
	  
    	    xdrmem_create (&xdres, xdrbuff, sizeof (procs_data) + MAXHOSTNAMELEN + 1, XDR_ENCODE);
	    xdr_procs_data (&xdres, &procsdata);
	    xdr_destroy (&xdres);
	    /* Send statistics to load master */
	    mon_sendmessage__ (sfd, &mid2, xdrbuff, 
			     sizeof (procs_data) + MAXHOSTNAMELEN + 1);
	    /* Free the allocated memory */
	    free (procsdata.hname);
	  }/* else the tid has been up too shortly and we didn't notice ... too bad */
	}
      }
    }
#endif


    // backup the old array for reference
#ifdef OSTYPE_linux
    for (int j=0; j<count_now; j++) {
      save_history[j].pid = New_save_history[j].pid;
      save_history[j].tid = New_save_history[j].tid;
      save_history[j].utime = New_save_history[j].utime;
      save_history[j].stime = New_save_history[j].stime;
      save_history[j].pcpu = New_save_history[j].pcpu;
      save_history[j].mem_res = New_save_history[j].mem_res;
    }
    max_procs = count_now;
    count_now = 0;
#endif

#ifdef USE_HASH_DATA

    /* copy the current watch_array to the old one ... */    
    for(int i=0;i<of_watch_tasks;i++)watch_array_old[i]=watch_array[i];
    of_watch_tasks_old=of_watch_tasks;

    /* now we set the alarm for mon_interval and start to wait for events on terminated tasks */

    req.tv_sec=1;
    req.tv_nsec=0;

    timer_expired=0;
    alarm(mon_interval);

#ifdef _DEBUG
    cerr<<"doing pvm_notify msg:1234 of_watch_tasks="<<of_watch_tasks<<endl;
#endif
    if(of_watch_tasks>0)pvm_notify(PvmTaskExit,1234,of_watch_tasks,watch_array);
    
    while(!timer_expired){
      /* we watch for termination messages from PVMD */
      while(pvm_probe(-1,1234)>0){
	/* receive the message */
        bufid=pvm_recv(-1,1234);
	/* there should be just integer describing the died tid */
	pvm_upkint(&died_tid,1,1);
	/* the process already died, so we can not consult /proc/$PID/stat to find out it's statistics,
	   so we will have to rely on what we collected so far ... */
	/* do we have something at all ??? */
	int i=0;
	int found=-1;
        while (i < max_procs) {
	  if(save_history[i].tid == died_tid) {
	    found = i;
	    i = max_procs;
	  }
	  i++;
	}
	/* if we have found some data on the died tid, just send it ... */
	if(found>=0){
	  
	  /* we already reported this !!!??? */
	  if(watch_array_old[found]<0)continue;
	  
	  /* fill in the values ... for now, just the last known we had */
	  procsdata.hname=strdup(hname);
	  procsdata.pid=save_history[found].pid;
	  procsdata.tid=save_history[found].tid;
	  
	  procsdata.pct_cpu=save_history[found].pcpu;
	  procsdata.pct_mem=0;
	  procsdata.mem_res=save_history[found].mem_res;
	  
	  /* always 1, since we monitor only moveable tasks */
	  procsdata.movable=1;
	  
	  /* Get timestamp */
	  if (gettimeofday(&ts, NULL) < 0) {
	    bailout ("mon_procs(), gettimeofday() failed");
	  }
	  /* We are only interested in seconds */
	  procsdata.ts = ts.tv_sec;
	  
    	  xdrmem_create (&xdres, xdrbuff, sizeof (procs_data) + MAXHOSTNAMELEN + 1, XDR_ENCODE);
	  xdr_procs_data (&xdres, &procsdata);
	  xdr_destroy (&xdres);
	  /* Send statistics to load master */
	  mon_sendmessage__ (sfd, &mid2, xdrbuff, 
			     sizeof (procs_data) + MAXHOSTNAMELEN + 1);
	  /* Free the allocated memory */
	  free (procsdata.hname);
	  
	  /* mark this entry as processed */
	  watch_array_old[found]=-1;
	  
	}/* else the tid has been up too shortly and we didn't notice ... too bad */
      }
      nanosleep(&req,NULL);
    }
#else /* USE_HASH_DATA */
    sleep(mon_interval);
#endif    
  } // while (1)
  
  /* Free the alocated memory */
  free (xdrbuff);
  free (hname);
  return (EXIT_SUCCESS);
} // main
