#ifndef MON_SLAVE_H
#define MON_SLAVE_H


#include <vector.h>

using namespace std;

/* Current cpuload */
typedef struct cpuload {
  double cl;           /* CPU load */
#ifndef OSTYPE_linux
  cpu_sysinfo_t ocl;   /* old CPU load */
#endif
  unsigned long idle;
  unsigned long wait;
  unsigned long user;
  unsigned long kernel;
} cpuload;


/* Complete node information */
typedef struct nodeinfo {
  long ncpu;               /* Number of CPUs */
#ifndef OSTYPE_linux
  processor_info_t pi;     /* Type of CPU */
#endif
  long psize;              /* Page size */
  long ppages;             /* Physical pages */
  long apages;             /* Available pages */
  long tram;               /* Total Physical RAM in kbytes */
  long uram;               /* Used phys. RAM in kbytes */
  long fram;               /* Free phys. RAM in kbytes */
  long tswap;              /* Total swap space installed [kB] */
  long fswap;              /* Free swap space [kB] */
  long uswap;              /* Used swap space [kB] */
#ifndef OSTYPE_linux
  cpu_stat_t cs;           /* CPU statistics */
#endif
  cpuload **bpcl;          /* Base pointer for CPU load */
  char **bpkscn;           /* Base pointer for ks_name of CPUs */
#ifndef OSTYPE_linux
  kstat_ctl_t *kctrl;
  kstat_t *klp;
#endif
  struct timeval ts;      /* Time stamp */
} nodeinfo;

void get_args(int argc, char *argv[], int *mon_interval);
int get_static_data(nodeinfo *ni);
int get_dynamic_data(nodeinfo *ni);


/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
/* used for calculating the cpu usage of each proc */
typedef struct save_hist {
  double pcpu;
  int pid;
  int tid; /* we have to remember this also to correctly identify died processes */
  long utime;
  long stime;
  long mem_res;	/* the resident memory size */
} save_hist;

#endif /* MON_SLAVE_H */
