#include <assert.h>

#include <stdlib.h>     
#include <stdio.h>     
#include <iostream.h>

#include <string.h>     /* for strlen */
#include <netdb.h>      /* for MAXHOSTNAMELEN */
#include <errno.h>      /* for errno */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>

#include <pthread.h>

#include "mon_global.h"
#include "mon_util.h"


int pvm_is_running()
{
  int uid;
  char pvmdpath[32];
  int success=0;
  struct stat sb;
  uid = (int) getuid();
  sprintf(pvmdpath,"/tmp/pvmd.%d",uid);
	     
  if (stat(pvmdpath, &sb) == -1) {
    if (errno == ENOENT)
    success = 0; /* not there yet */
  } else {
    success=1;
  }
  return success;
}
					     
					     

/* Encode / Decode message type id */
int xdr_mtype_id (XDR *xdrs,  message_id *id)
{
  if (!xdr_int(xdrs, ((int *) &id->mtype))) {
    return FALSE;
  }
  return TRUE;
}


/* Encode / Decode load monitor data */
int xdr_lm_data (XDR *xdrs, lm_data *data)
{  
  if (!xdr_string(xdrs,  &(data->hname), MAXHOSTNAMELEN+1)) {
    fprintf (stderr, "xdr_lm_data(), xdr_string() failed\n");
    return FALSE;
  }

  if (!xdr_int (xdrs, &data->hid)) {
    fprintf (stderr, "xdr_lm_data(), xdr_int() for hid failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->ncpu)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for ncpu failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->cpuload)) {
    fprintf (stderr, "xdr_lm_data(), xdr_double() for cpuload failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->trend)) {
    fprintf (stderr, "xdr_lm_data(), xdr_double() for trend failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->uram)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for uram failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->fram)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for fram failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->uswap)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for uswap failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->fswap)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for fswap failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->ts)) {
    fprintf (stderr, "xdr_lm_data(), xdr_long() for ts failed\n");
    return FALSE;
  }

  return TRUE;
}


/* Encode / Decode process monitor data */
int xdr_procs_data (XDR *xdrs, procs_data *data)
{  
  if( !xdr_int (xdrs, &data->movable )) {
    fprintf (stderr,
	     "xfr_procs_data(), xdr_int() for movable failed\n" );
    return FALSE;
  }


  if (!xdr_int (xdrs, &data->hid)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_int() for hid failed\n");
    return FALSE;
  }

  if (!xdr_int (xdrs, &data->tid)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_int() for tid failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->pid)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_long() for pid failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->pct_cpu)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_double() for pct_cpu failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->pct_mem)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_double() for pct_mem failed\n");
    return FALSE;
  }

  if (!xdr_u_long (xdrs, &data->mem_img)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_u_long() for mem_img failed\n");
    return FALSE;
  }

  if (!xdr_u_long (xdrs, &data->mem_res)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_u_long() for mem_res failed\n");
    return FALSE;
  }

  if (!xdr_long (xdrs, &data->ts)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_long() for ts failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->totals)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_double() for totals failed\n");
    return FALSE;
  }

  if (!xdr_double (xdrs, &data->total_pct_cpu)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_double() for total_pct_cpu failed\n");
    return FALSE;
  }

  char *addr=data->pname;
  if (!xdr_string(xdrs,  &(addr), 128+1)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_string() for pname failed\n");
    return FALSE;
  }


  if (!xdr_string(xdrs,  &(data->hname), MAXHOSTNAMELEN+1)) {
    fprintf (stderr, 
	     "xdr_procs_data(), xdr_string() for hname failed\n");
    return FALSE;
  }

  return TRUE;
}

resdes *new_resdes_struct()
{
  resdes *rd;
  
  /* Allocate memory for the entry */
  if ((rd = (resdes *) calloc (sizeof (resdes), sizeof (char))) == NULL) {
    bailout ("new_resdes_struct(), calloc() failed");
  }
  return rd;
}


/* resdes **rd_addr;   Pointer to base address of resource entry */
/* Returns number of successfully  read entries  */
int read_resdes (resdes **rd_addr, char* Filename) 
{

  resdes *tmp_rd;  /* Temporary pointer to resource entry */
  FILE *f;         /* Resource description file descriptor */
  char *line;      /* Line read in */
  int n = 0;       /* Successful lines read so far */


  /* Allocate memory to read in a line */
  if ((line = (char *) calloc (DEFSTR, sizeof (char))) == NULL) {
    fprintf (stderr, "read_res_descr(), calloc() failed\n");
    return (-1);
  }
  
  /* Open the resource description file for reading */
  if ((f = fopen (Filename, "r")) == NULL) {
    fprintf (stderr, "read_res_descr(), fopen(\"%s\") failed\n", Filename); 
    return (-1);
  }
  
  /* Find section [hosts] */
  while ((fgets (line, DEFSTR, f)) != NULL ) {
    if( strncmp( line, "[hosts]", strlen("[hosts]" )) == 0 )	break;
  }
  if( strncmp( line, "[hosts]", strlen("[hosts]" )) != 0 ) {
	bailout( "read_resdes(): unable to find [hosts] section\n" );
  }

  /* Read in the lines */
  int HostId = 0;
  while ((fgets (line, DEFSTR, f)) != NULL ) {

    /* stop, if new section starts */
    if( line[0] == '[' )	break;

    /* Ignore comments */
    if (line[0] != '#') {

      /* First line ever ? */
      if (*rd_addr == NULL) {

	/* Allocate memory for the entry */
	*rd_addr = new_resdes_struct();
	tmp_rd = *rd_addr;

      } else {

	tmp_rd->next = new_resdes_struct();
	tmp_rd = tmp_rd->next;
      }

      /* Seperate into tokens */
      tmp_rd->hid = ++HostId;
      char* p = strtok( line, " \t" );
      assert( p );
      strcpy( tmp_rd->hname, p );
      p = strtok( 0, " \t" );
      assert( p );
      strcpy( tmp_rd->ip, p );
      p = strtok( 0, " \t" );
      if( p ) {
          tmp_rd->cpu_weight = atof( p );
          p = strtok( 0, " \t" );
      } else {
          tmp_rd->cpu_weight = 1;
      }
      if( p ) {
          tmp_rd->mem_weight = atof( p );
      } else {
          tmp_rd->mem_weight = 1;
      }
      if( tmp_rd->mem_weight <= 0 )     tmp_rd->mem_weight = 1;
      if( tmp_rd->cpu_weight <= 0 )     tmp_rd->cpu_weight = 1;
      n++;
    }
  }
  return (n);
}  


int get_resdes_hid (resdes *rd_base, char *hname) 
{
  resdes *tmp_rd;
  
  /* Determine host id of this host from the resource entries */
  for (tmp_rd = rd_base; tmp_rd != NULL; tmp_rd = tmp_rd->next) {
#ifdef _DEBUG
    cerr<<hname<<" == "<<tmp_rd->hname<<endl;
#endif
    if (strcmp (hname, tmp_rd->hname) == 0) {
      /* Found the string */
      return (tmp_rd->hid);
    } else if( strchr( hname, '.' ) != 0  &&  strchr( tmp_rd->hname, '.' ) == 0 ) {
        char* dot = strchr( hname, '.' );
        *dot = '\0';
        if( strcmp( hname, tmp_rd->hname ) == 0 ) {
            *dot = '.';
            return tmp_rd->hid;
        }
        *dot = '.';
    } else if( strchr( hname, '.' ) == 0  &&  strchr( tmp_rd->hname, '.' ) != 0 ) {
        char* dot = strchr( tmp_rd->hname, '.' );
        *dot = '\0';
        if( strcmp( hname, tmp_rd->hname ) == 0 ) {
            *dot = '.';
            return tmp_rd->hid;
        }
        *dot = '.';
    }
  }
  
  return (-1);
}


/* Returns true, if any options were read   */
int read_options (t_global_options *options, char* Filename)
{

  FILE *f;         /* Resource description file descriptor */
  char *line;      /* Line read in */
  char *key;       /* option to be set */
  char *value;     /* value of that option */
  int n = 0;       /* Successful lines read so far */
  int StartSlaveCmdSet = 0;
  int DumpDirSet = 0;

  /* Allocate memory to read in a line */
  if ((line = (char *) calloc (DEFSTR, sizeof (char))) == NULL) {
    fprintf (stderr, "read_options(), calloc() failed\n");
    return (-1);
  }

  /* Open the resource description file for reading */
  if ((f = fopen (Filename, "r")) == NULL) {
    fprintf (stderr, "read_options(), fopen(\"%s\") failed\n", Filename);
    return (-1);
  }

  /* Find section [options] */
  while ((fgets (line, DEFSTR, f)) != NULL ) {
    if( strncmp( line, "[options]", strlen("[options]" )) == 0 )    break;
  }
  if( strncmp( line, "[options]", strlen("[options]" )) != 0 ) {
    bailout( "read_options(): unable to find [options] section\n" );
  }

  /* Make some initializations */
  options->CPUweight = 1;
  options->MEMweight = 0;
  options->MIGweight = 0.3;
  options->MaxSearchTime = 60;

  /* Read in the lines */
  while ((fgets (line, DEFSTR, f)) != NULL ) {

    /* stop, if new section starts */
    if( line[0] == '[' )        break;

    /* Ignore comments */
    if (line[0] != '#') {
      key = strtok( line, " \t=\n" );
      if( !key )  continue;
      if( strcmp( key, "START_SLAVE_COMMAND" ) == 0 ) {
        value = strtok( 0, "\n \t=" );
        if( value )  strcpy( options->StartSlaveCmd, value );
	else         bailout( "read_options(): illegal format for START_SLAVE_COMMAND" );
	while( value=strtok( 0, " \t\n" ) ) {
		strcat( options->StartSlaveCmd, " " );
		strcat( options->StartSlaveCmd, value );
        }
	StartSlaveCmdSet = 1;
      } else if( strcmp( key, "DUMP_DIR" ) == 0 ) {
	value = strtok( 0, "\n\t =" );
	if( value )  strcpy( options->DumpDir, value );
	else         bailout( "read_options(): illegal format for DUMP_DIR" );
	DumpDirSet = 1;
      } else if( strcmp( key, "CPU_WEIGHT" ) == 0 ) {
	value = strtok( 0, "\n\t =" );
	if( value )	options->CPUweight = atof( value );
	else		bailout( "read_options(): illegal format for CPU_WEIGHT" );
      } else if( strcmp( key, "MEM_WEIGHT" ) == 0 ) {
        value = strtok( 0, "\n\t =" );
        if( value )     options->MEMweight = atof( value );
        else            bailout( "read_options(): illegal format for MEM_WEIGHT" );
      } else if( strcmp( key, "MIG_WEIGHT" ) == 0 ) {
        value = strtok( 0, "\n\t =" );
        if( value )     options->MIGweight = atof( value );
        else            bailout( "read_options(): illegal format for MIG_WEIGHT" );
      } else if( strcmp( key, "MAX_SEARCH_SECS" ) == 0 ) {
        value = strtok( 0, "\n\t =" );
        if( value )     options->MaxSearchTime = atoi( value );
        else            bailout( "read_options(): illegal format for MAX_SEARCH_TIME" );
      } else {
	fprintf( stderr, "read_options(): unknown option \"%s\"\n", key );
	exit( -1 );
      }
    }
  }

  if( !StartSlaveCmdSet )	bailout( "read_options(): START_SLAVE_COMMAND not set in config file." );
  if( !DumpDirSet )	bailout( "read_options(): DUMP_DIR not set in config file." );

  fclose(f);
  return n;
}

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
char *skip_token(const char *p)
{
    while (isspace(*p)) p++;
    while (*p && !isspace(*p)) p++;
    return (char *)p;
}

/* for LinuX, added by streit@upb.de at 22th Nov 1999 */
char *skip_ws(const char *p)
{
    while (isspace(*p)) p++;
    return (char *)p;
}

char *read_procname(const char *p,char *name){
    while ((!isspace(*p))&&(*p!=0)&&(*p!=')')){
      *name=*p;
      name++;
      p++;
    }
    if(*p==')')p++;
    *name=0;
    return (char *)p;
}


// Like write(), but either sends all or nothing at all
size_t writeall( int fd, const void* buf, size_t nbyte )
{
    size_t BytesSent = 0;
    while( nbyte > BytesSent ) {
	size_t Result = write( fd, ((char*)buf)+BytesSent, nbyte-BytesSent );
	if( Result < 0 )    return Result;
        BytesSent += Result;
    }
    return BytesSent;
}


