#include <pthread.h>
#include <assert.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdlib.h>
#include <stdio.h>

#include <rpc/types.h>
#include <rpc/xdr.h>

#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <unistd.h>

#include <string.h>
#include <strings.h>

#include <pthread.h>
#include <signal.h>

#include <string.h>
#include <vector.h>
//#include <iostream.h>

#include "pvm3.h"

#include "monitor.h"
#include "migrator.h"
#include "mon_global.h"
#include "balancer.h"

#include <errno.h>

using namespace std;

/***********************************************/
/*****  Some global variables ******************/
/***********************************************/

resdes *rd_base = NULL;   /* Base pointer to resource description */

char *montypes[] = {"SEND", "RECV"};

/* Number of successfully read resource entries */
int rd_entries;

/* The load queue */
load_queue loadq;

/* Path name of the config file */
char* configfile;

/* The communication queue */
comm_queue commq;

/* The procs queue */
procs_queue procsq;
 
/* Base pointer to array of pointers to load statistics of each host */
load_stats **lm_stats_base;

/* Base pointer to process statistics of each host */
procs_hosts_stats *pd_hosts_stats;


/* Mutex and condition variable for the load queue */
pthread_mutex_t lq_mutex; 
pthread_cond_t  lq_cond;  


/* Mutex and condition variable for the communication queue */
pthread_mutex_t cq_mutex; 
pthread_cond_t  cq_cond;  


/* Mutex and condition variable for the procs queue */
pthread_mutex_t pq_mutex; 
pthread_cond_t  pq_cond;  

/* Mutex for accessing the LOAD statistics (scheduler) */
pthread_mutex_t ls_mutex;

/* Mutex for accessing the PROCS statistics (scheduler) */
pthread_mutex_t ps_mutex;

/* Mutex for accessing the HISTORY statistics (scheduler) */
pthread_mutex_t hs_mutex;

/* Global PVM task id */
int tid;

/* Task id of the local PVMD */
int dtid;

/***********************************************/
/*****************************************************/
static
int
read_all( unsigned long socket, char* buffer, unsigned Bytes2Read )
{
    // Make sure that exactly Bytes2Read bytes are received
    int BytesReceived = 0;
    while( Bytes2Read > BytesReceived ) {
        int Result = recv( socket, buffer+BytesReceived,
                           Bytes2Read-BytesReceived, 0 );
        if( Result <= 0 )    return Result;
        BytesReceived += Result;
    }
    return BytesReceived;
}




/***********************************************/
void
ThreadIgnoreSigAlarm()
{
    sigset_t set;
    sigemptyset( &set );
    sigaddset( &set, SIGALRM );
    pthread_sigmask( SIG_BLOCK, &set, 0 );
}




/***********************************************/

t_global_options global_options;
int ConnectedToPVM = 0;

void startup_slaves(resdes *rd_base, char* ResourceFile)
{
  pid_t cpid;         /* Child process ID */
  int cstatus;        /* Child exit status */
  static char ip[16];        /* IP number where to spawn load monitor slave */
  resdes *tmp_rd;
  
  for (tmp_rd = rd_base; tmp_rd != NULL; tmp_rd = tmp_rd->next) {

    if ((cpid = fork ()) < 0) {
      bailout ("startup_slaves(), fork() failed");

    } else {

#ifdef _DEBUG
      cerr<<getpid()<<"_debug: child pid "<<cpid<<endl;
#endif

      if (cpid == 0) {

	/* This is the first child process, fork again to */
	/* prevent zombie state */
	if ((cpid = fork ()) < 0) {
	  bailout ("startup_slaves(), second fork() failed");

	} else {

	  if (cpid > 0) {
#ifdef _DEBUG
 	    cerr<<getpid()<<"_debug: first child pid "<<cpid<<" terminated"<<endl;
#endif



	    // Terminate first child which is parent of second child
	    exit (EXIT_SUCCESS);

	  } else {
#ifdef _DEBUG
 	    cerr<<getpid()<<"_debug: second child pid "<<cpid<<endl;
#endif


	    /* This is the second child, wait for termination of parent */
	    sleep(2);

	    /* Spawn remote load monitor */
            char PortAsString[40];
            sprintf( PortAsString, "%d", global_options.MonPort );
#ifdef _DEBUG
	    cerr<<getpid()<<"_debug: start mon_slave on IP_"<<tmp_rd->ip<<"_"<<endl;
	    cerr<<getpid()<<"_debug: Cmd: "<<global_options.StartSlaveCmd<<endl
		<<getpid()<<"_debug: MonHost: "<<global_options.MonHost<<endl
		<<getpid()<<"_debug: PortAsString: "<<PortAsString<<endl
		<<getpid()<<"_debug: global_options.MonPort: "<<global_options.MonPort<<endl
		<<getpid()<<"_debug: resource file: "<<ResourceFile<<endl;
	    cerr<<getpid()<<"_debug: rcmd: "<<RSHCMD<<endl;
#endif
 	    cerr<<"doing it now ...";
	    if (execlp (RSHCMD, tmp_rd->ip,
			global_options.StartSlaveCmd, 
			global_options.MonHost,
			PortAsString,
			ResourceFile,
 			NULL) < 0) {
	      bailout ("startup_slaves(), execlp() failed");
	    }
	    exit (EXIT_SUCCESS);
	  }
	}

      } else {

	/* This is the parent process */
	/* Wait for termination of first child process */
	if (waitpid (cpid, NULL, 0) != cpid) {
	  bailout ("startup_slaves(), waitpid() failed");
	}
      }
    }
  }
}



int xdr_timespec(struct XDR* xdrs, struct timespec* tsp)
{
	if (!xdr_long(xdrs, &tsp->tv_sec))
		return FALSE;
	if (!xdr_long(xdrs, &tsp->tv_nsec))
		return FALSE;
	return TRUE;
}




int
xdr_mondata(struct XDR* xdrs, struct mon_data*mdp)
{
	if (!xdr_int(xdrs, &mdp->mon_montype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_msgtype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_src))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_dst))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_len))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_stime))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_etime))
		return FALSE;
	return TRUE;
}




int master_socket_init()
{
  int msfd;                   /* Master socket file descriptor */
  int rsa = 1;                /* Reuse local socket address */
  struct sockaddr_in msds;    /* Master socket data structure */
  
  /* Create master socket */
  msfd = socket(AF_INET, SOCK_STREAM, 0);
  if (msfd < 0) {
    bailout ("master_socket_init(), socket() failed");
  }
#ifdef _DEBUG
  cerr<<getpid()<<"_debug: socket "<<msfd<<endl;
#endif

  /* Reuse socket */
  if (setsockopt(msfd, 
		 SOL_SOCKET, 
		 SO_REUSEADDR, 
		 (char *) &rsa, 
		 sizeof (rsa)) < 0) {
    bailout ("master_socket_init(), setsockopt() failed");
  }
  
  msds.sin_family = AF_INET;
#ifdef HAS_SINLEN
  msds.sin_len = sizeof(struct sockaddr_in);
#endif
  msds.sin_addr.s_addr = htonl(INADDR_ANY);
  msds.sin_port = htons(0);

  /* Bind the master socket to the port */
  if (bind (msfd, (struct sockaddr *) &msds, sizeof(msds)) != 0) {
    bailout ("master_socket_init(), bind() failed");
  }
#ifndef OSTYPE_linux  
  int length = sizeof( msds );
#else
  socklen_t length = sizeof(msds);
#endif

  if( getsockname( msfd, (struct sockaddr*)&msds,(socklen_t*) &length) != 0 ) {
            bailout( "master_socket_init(), getsockname() failed" );
  }
  global_options.MonPort = ntohs( msds.sin_port );

  /* Set the master monitor's hostname in global_options */
  if( !getenv("HOST") ) {
      bailout( "master_socket_init(), cannot determine hostname" );
  } else {
      strcpy( global_options.MonHost, getenv( "HOST" ) );
      cout << "Master monitor resides on host \"" << global_options.MonHost 
           << "\" port " << global_options.MonPort << endl;
  }

  /* Listen on master socket */
  if (listen (msfd, 5) != 0) {
    bailout ("master_socket_init(), listen() failed");
  }
  return msfd;
}



void process_LOADM_message(lm_data *mb) 
{
  ll_load_queue *ll_lq;

  
  /* Allocate memory */
  if ((ll_lq = (ll_load_queue *) calloc (sizeof (ll_load_queue), 
					 sizeof (char))) == NULL) {
    bailout ("process_LOADM_message(), calloc failed");
  }

  /* Copy the data into it */
  ll_lq->lmd = *mb;
  ll_lq->next = NULL;
  
  /* Try to aquire the mutex for the load queue */
  pthread_mutex_lock (&lq_mutex);

  /* Append to the load queue */
  if (loadq.tail == NULL) {

    /* We are appending to an empty queue */
    loadq.tail = ll_lq;
    loadq.head = ll_lq;

  } else {
  
    /* There is already an entry in the queue, append to this one */
    loadq.tail->next = ll_lq;

    /* Update tail of the queue */
    loadq.tail = ll_lq;
  }

  /* Send a signal to the load thread that there is new data in the queue */
  pthread_cond_signal (&lq_cond);

  /* Free the mutex for the load queue */
  pthread_mutex_unlock (&lq_mutex);
}



void process_COMMM_message(struct mon_data *md)
{
  ll_comm_queue *ll_cq;

  
  /* Allocate memory */
  if ((ll_cq = (ll_comm_queue *) calloc (sizeof (ll_comm_queue), 
					 sizeof (char))) == NULL) {
    bailout ("process_COMM_message(), calloc failed");
  }
  
  /* Copy the data into it */
  ll_cq->cmd = *md;
  ll_cq->next = NULL;
  
  /* Try to aquire the mutex for the communication queue */
  pthread_mutex_lock (&cq_mutex);

  /* Append to the communication queue */
  if (commq.tail == NULL) {

    /* We are appending to an empty queue */
    commq.tail = ll_cq;
    commq.head = ll_cq;

  } else {
  
    /* There is already an entry in the queue, append to this one */
    commq.tail->next = ll_cq;

    /* Update tail of the queue */
    commq.tail = ll_cq;
  }

  /* Send a signal to the comm. thread that there is new data in the queue */
  pthread_cond_signal (&cq_cond);

  /* Free the mutex for the communication queue */
  pthread_mutex_unlock (&cq_mutex);
}


void process_PROCM_message(procs_data *mb,MTYPE tp) 
{
  ll_procs_queue *ll_pq;

  
  /* Allocate memory */
  if ((ll_pq = (ll_procs_queue *) calloc (sizeof (ll_procs_queue), 
					  sizeof (char))) == NULL) {
    bailout ("process_PROCM_message(), calloc failed");
  }

  /* Copy the data into it */
  ll_pq->pmd = *mb;
  ll_pq->next = NULL;
  
  /* if this is a termination message, set movable to -2 */
  if(tp==TERMS){
#ifdef _DEBUG
    cerr<<"Intercepted a termination packet for tid="<<ll_pq->pmd.tid<<endl;
#endif
    ll_pq->pmd.movable=-2;
  }
  
  /* Try to aquire the mutex for the procs queue */
  pthread_mutex_lock (&pq_mutex);

  /* Append to the procs queue */
  if (procsq.tail == NULL) {

    /* We are appending to an empty queue */
    procsq.tail = ll_pq;
    procsq.head = ll_pq;

  } else {
  
    /* There is already an entry in the queue, append to this one */
    procsq.tail->next = ll_pq;

    /* Update tail of the queue */
    procsq.tail = ll_pq;
  }

  /* Send a signal to the procs thread that there is new data in the queue */
  pthread_cond_signal (&pq_cond);

  /* Free the mutex for the procs queue */
  pthread_mutex_unlock (&pq_mutex);
}




void get_messages(int msfd)
{
  fd_set rfds, 
         efds,
         tmprfds,
         tmpefds;             /* Reading and exception file descriptor sets */
  int maxfd;                  /* Highest socket descriptor */
  int fdready;                /* Number of fd ready for reading */
  int newfd;                  /* Socket fd from accept() */

  struct mon_data commb;      /* Message buffer for communication messages */
  lm_data loadb;              /* Message buffer for load messages */
  procs_data procsb;          /* Message buffer for process stat. messages */
  XDR xdrds;                  /* XDR decode stream */
  char *xdrbuff;              /* XDR buffer */
  int xdrbs;                  /* XDR buffer size */

  message_id mb_id;           /* Message buffer for message ident. */
  XDR xdrds_id;               /* XDR decode stream for message ident. */
  char *xdrbuff_id;           /* XDR buffer for message identification */
  int xdrbs_id;               /* XDR buffer size for message ident. */

  int bytes2read;             /* Data message size to be read from socket */
  int n;                      /* Bytes read from socket */
  int i;
  
  FD_ZERO(&rfds);
  FD_ZERO(&efds);
  
  FD_SET(msfd, &rfds);
  FD_SET(msfd, &efds);
  
  maxfd = msfd;

  /* Allocate XDR buffer for the message identification */
  xdrbs_id = sizeof (message_id);
  if ((xdrbuff_id = (char *) calloc (xdrbs_id, sizeof (char))) == NULL) {
    bailout ("get_messages(), calloc() for xdrbuff_id failed");
  }
    
  /* Allocate XDR buffer for received messages */
  xdrbs = DEFRBUF;
  if ((xdrbuff = (char *) calloc (xdrbs, sizeof (char))) == NULL) {
    bailout ("get_messages(), calloc() for xdrbuff failed");
  }

  while (1) {

    /* Save fd sets */
    tmprfds = rfds;
    tmpefds = efds;
    
    /* Wait for connections */
    if ((fdready = select (maxfd + 1, &tmprfds, NULL, &tmpefds, NULL)) < 0) {
      bailout ("get_messages(), select() failed");
    }
#ifdef _DEBUG
    cerr<<getpid()<<"_monitor: got a new message\n";
#endif
    /* Check for new connection on master socket */
    if (FD_ISSET(msfd, &tmprfds)) {

      /* Accept new connection */
      if ((newfd = accept (msfd, 0, 0)) < 0) {
	bailout ("get_messages(), accept() failed");
      } else {
        cerr << "New connection established" << endl;
	/* Add new fd */
	FD_SET (newfd, &rfds);
	FD_SET (newfd, &efds);
	/* New fd higher as current max */
	if (newfd > maxfd) {
	  maxfd = newfd;
	}
	/* Clear master socket bit */
	FD_CLR(msfd, &tmprfds);
	FD_CLR(msfd, &tmpefds);
      }
    }

    /* Check the remainig descriptors */
    for (i = 0; i <= maxfd; i++) {

      /* Check for exceptions */
      if (FD_ISSET(i, &tmpefds)) {

        cerr << ">>> Slave disconnected 1 <<<" << endl;
	/* Remove socket form fd set */
	FD_CLR(i, &rfds);
	FD_CLR(i, &efds);
	shutdown(i, 2);
	close(i);
	if (i == maxfd) {
	  maxfd--;
	  continue;
	}
      }
      if (FD_ISSET(i, &tmprfds)) {

	/* Try to read the message identification */
	if ((n = read_all(i, xdrbuff_id, xdrbs_id)) <= 0) {

	  /* Error reading from socket, or EOF received close it */
          cerr << ">>> Slave disconnected 2 <<<" << endl;
#ifdef _DEBUG
	  perror("Slave disconnected: ");
#endif
	  FD_CLR(i, &rfds);
	  FD_CLR(i, &efds);
	  shutdown(i, 2);
	  close(i);
	  if (i == maxfd) {
	    maxfd--;
	    continue;
	  }
	} else {

          assert( n==xdrbs_id );

	  /* Decode the message identification */
	  xdrmem_create (&xdrds_id, xdrbuff_id, xdrbs_id, XDR_DECODE);
	  if (xdr_mtype_id (&xdrds_id, &mb_id) != TRUE) {
	    bailout ("get_messages(), xdr_mtype_id() failed");
	  }
	  xdr_destroy (&xdrds_id);

	  /* Set correct size of the data message */ 
	  switch (mb_id.mtype) {

	  case LOADM:
	    bytes2read = sizeof (lm_data) + MAXHOSTNAMELEN + 1;
	    break;
	    
	  case COMMM:
	    bytes2read = sizeof (struct mon_data);
	    break;

	  case PROCM:
	    bytes2read = sizeof (procs_data) + MAXHOSTNAMELEN + 1;
	    break;

	  case TERMS:
	    bytes2read = sizeof (procs_data) + MAXHOSTNAMELEN + 1;
	    break;

	  default:
              cerr << "### n=" << n << " tag=" << mb_id.mtype << endl;
              bailout( "get_messages(), received unknown message" );
	  }
	    
	  /* Try to read the message data*/
	  if ((n = read_all(i, xdrbuff, bytes2read)) <= 0) {
	    /* Error reading from socket, or EOF received close it */
            cerr << ">>> Slave disconnected 3 <<<" << endl;
	    FD_CLR(i, &rfds);
	    FD_CLR(i, &efds);
	    shutdown(i, 2);
	    close(i);
	    if (i == maxfd) {
	      maxfd--;
	      continue;
	    }
	  }

	  /* Process the data message according to 
	     the message identification tag */ 
#ifdef _DEBUG
	  cerr<<getpid()<<"_message type: "<<mb_id.mtype<<endl;
#endif
	  switch (mb_id.mtype) {
	    
	  case LOADM:
	    /* Clear message buffer for load messages */
	    bzero (&loadb, sizeof (lm_data));

	    /* Decode received data */
	    xdrmem_create (&xdrds, xdrbuff, bytes2read, XDR_DECODE);
	    if (xdr_lm_data (&xdrds, &loadb) != TRUE) {
	      bailout ("get_messages(), xdr_lm_data() failed");
	    }
	    xdr_destroy (&xdrds);

	    /* Process the buffer */
	    process_LOADM_message(&loadb);
	    break;

	  case COMMM:
	    /* Clear message buffer for communication messages */
	    bzero (&commb, sizeof (struct mon_data));

	    /* Decode received data */
	    xdrmem_create (&xdrds, xdrbuff, bytes2read, XDR_DECODE);
	    if (xdr_mondata (&xdrds, &commb) != TRUE) {
	      bailout ("get_messages(), xdr_mondata() failed");
	    }
	    xdr_destroy (&xdrds);

	    /* Process the buffer */
	    process_COMMM_message(&commb);
	    break;

	  case PROCM:
	  case TERMS:
	    /* Clear message buffer for process statistics messages */
	    bzero (&procsb, sizeof (procs_data));

	    /* Decode received data */
	    xdrmem_create (&xdrds, xdrbuff, bytes2read, XDR_DECODE);
	    if (xdr_procs_data (&xdrds, &procsb) != TRUE) {
	      bailout ("get_messages(), xdr_procs_data() failed");
	    }
#ifdef _DEBUG
            cerr << "** Xdr receive " << procsb.tid << " movable == " << procsb.movable << endl;
#endif

	    xdr_destroy (&xdrds);

	    /* Process the buffer */
	    process_PROCM_message(&procsb,mb_id.mtype);
	    break;
	  }
#ifdef _DEBUG
	  cerr<<getpid()<<"_successfully handle the message\n";
#endif
	}
      }
    }
  } // while
}





int main(int argc, char *argv[])
{
#ifdef _DEBUG
  cerr<<"hi, im monitor with pid "<<getpid()<<endl;
#endif
  resdes *tmp_rd;

  pthread_attr_t pattr1, pattr2, pattr3, pattr4, pattr5;
  pthread_t      thr1, thr2, thr3, thr4, thr5;

  int msfd;                   /* Master file descriptor */

  /* Check command line options */
  if( argc != 2 ) {
	bailout( "usage: monitor config_file\n" );
  }

  /* Read options from config file */
  configfile = argv[1];
  read_options( &global_options, argv[1] );

  /* Read resource description */
  if ((rd_entries = read_resdes (&rd_base, argv[1])) < 1) {
    bailout ("main(), can't parse resource description");
  }

  /* Initialize thread objects with default values */
  pthread_attr_init (&pattr1);
  pthread_attr_init (&pattr2);
  pthread_attr_init (&pattr3);
  pthread_attr_init (&pattr4);
  pthread_attr_init (&pattr5);

  pthread_cond_init (&lq_cond, NULL);
  pthread_cond_init (&cq_cond, NULL);
  pthread_cond_init (&pq_cond, NULL);

  pthread_mutex_init (&lq_mutex, NULL);
  pthread_mutex_init (&cq_mutex, NULL);
  pthread_mutex_init (&pq_mutex, NULL);

  pthread_mutex_init (&ls_mutex, NULL);
  pthread_mutex_init (&ps_mutex, NULL);
  pthread_mutex_init (&hs_mutex, NULL);

  /* Start thread for monitoring PVM applications */
  pthread_create (&thr1, &pattr1, process_COMM_queue, NULL);

  /* Start thread for monitoring node load */
  pthread_create (&thr2, &pattr2, process_LOAD_queue, NULL);

  /* Start thread for monitoring process statisitics */
  pthread_create (&thr3, &pattr3, process_PROCS_queue, NULL);

  /* Start scheduler */
  pthread_create (&thr4, &pattr4, scheduler, NULL);

  /* Start pvm connector */
  pthread_create (&thr5, &pattr5, pvm_connect_thr, NULL );

  ThreadIgnoreSigAlarm();

  /* Initialize master socket */
  if ((msfd = master_socket_init()) < 0) {
    bailout ("master_socket_init() failed");
  }

#ifdef _DEBUG
  cerr<<getpid()<<"_debug: master socket: "<<msfd<<endl;
  cerr<<getpid()<<"_debug: this is monitor "<<getpid()
      <<" before starting the slaves\n";
#endif

  /* Start slave monitor tasks */
  startup_slaves(rd_base, argv[1] ); 
#ifdef _DEBUG
  cerr<<getpid()<<"_debug: this is monitor "<<getpid()
      <<" after starting the slaves\n";
#endif
  /* Process all incoming messages */
  get_messages(msfd);
  
  return 0;
}

int copy_LOAD_data (lm_data *to, lm_data *from)
{
  lm_data *tmp_ptr;
  
  /* Copy the data */
  memcpy (to, from, sizeof (lm_data));
  if ((to->hname = strdup (from->hname)) == NULL) {
    fprintf (stderr, "copy_LOAD_data(), strdup() failed\n");
    return (-1);
  }

  return (0);
}


/* Currently implemented is the mean average of a time window */
void update_LOAD_stats (load_stats *node, ll_load_queue *qdata)
{
  double d;
  double k;
  int    m;
  
  /* New statistics? */
  if (node->ticks == 0) {

    /* Initialize ring buffer */
    if ((node->rb_head = 
	 (lm_data *) calloc (DEFMONWT, sizeof (lm_data))) == NULL) {
      bailout ("update_LOAD_stat(), calloc() for rb_head failed");
    }
    node->rb_tail = node->rb_head + DEFMONWT - 1;
    node->rb_pos  = node->rb_head;
    
    /* Preserve a copy of the data */
    if (copy_LOAD_data (node->rb_pos, &(qdata->lmd)) < 0) {
      bailout ("update_LOAD_stat(), copy_LOAD_data() for node->rb_pos failed");
    }

    /* Initialize statistics */
    memcpy (&(node->stats), &(qdata->lmd), sizeof (lm_data));
    if ((node->stats.hname = strdup (qdata->lmd.hname)) == NULL) {
      bailout ("update_LOAD_stat(), strdup() for node->stats.hname failed");
    }
    
    node->rb_pos++;
    node->ticks++;

    return;

  } else {
    
    /* Ticks for time window complete? */
    if (node->ticks == DEFMONWT) {

      if (node->rb_pos > node->rb_tail) {
	/* We would write behind the array, correct this */
	node->rb_pos = node->rb_head;
      }

      /* _           _         1
       * x(t=t   ) = x(t=t ) - - ( x(t=t - nT) - x(t=t   ))
       *      j+1         j    n        j             j+1
       */

      k = 1.0 / (double) DEFMONWT;    /* 1/n */

      node->stats.cpuload -= k * (node->rb_pos->cpuload - qdata->lmd.cpuload);

      /* Attention signed long ahed, negativ values are not good ... */
      d = (double) node->stats.uram;
      d -= k * (((double) node->rb_pos->uram)  - ((double) qdata->lmd.uram));
      d = (d < 0) ? 0 : d;
      node->stats.uram = (long) d;

      d = (double) node->stats.fram;    
      d -= k * (((double) node->rb_pos->fram)  - ((double) qdata->lmd.fram));
      d = (d < 0) ? 0 : d;
      node->stats.fram = (long) d;
      
      d = (double) node->stats.uswap;
      d -= k * (((double) node->rb_pos->uswap) - ((double) qdata->lmd.uswap));
      d = (d < 0) ? 0 : d;
      node->stats.uswap = (long) d;

      d = (double) node->stats.fswap;
      d -= k * (((double) node->rb_pos->fswap) - ((double) qdata->lmd.fswap));
      d = (d < 0) ? 0 : d;
      node->stats.fswap = (long) d;

      node->stats.ts       = qdata->lmd.ts;
      node->stats.trend	   = qdata->lmd.trend;

      /* Check values for < 0 to prevent -0.000000... */
      node->stats.cpuload = 
	(node->stats.cpuload < 0) ? 0 : node->stats.cpuload;
 
      /* Free the dynamic data of the current entry, its obsolete */
      free (node->rb_pos->hname);

      /* Preserve a copy of the new one */
      if (copy_LOAD_data (node->rb_pos, &(qdata->lmd)) < 0) {
	bailout ("update_LOAD_stats(), copy_LOAD_data() for node->rb_pos failed");
      }
      
      node->rb_pos++;
      
      return;

    } else {

      /* Preserve a copy of the data */
      if (copy_LOAD_data (node->rb_pos, &(qdata->lmd)) < 0) {
	bailout ("update_LOAD_stat(), copy_LOAD_data() for node->rb_pos failed");
      }
      
      /* _         1         _
       * x(t=t ) = - ( (j-1) x(t=t   ) + x(t=t ) )
       *      j    j              j-1         j
       */
      
      k = 1.0 / (double) (node->ticks + 1);  /* 1/j */
      m = node->ticks;                       /* j-1 */
      
      node->stats.cpuload = k * (m * node->stats.cpuload + qdata->lmd.cpuload);
      node->stats.uram    = (long)(k * (m * node->stats.uram    + qdata->lmd.uram));
      node->stats.fram    = (long)(k * (m * node->stats.fram    + qdata->lmd.fram));
      node->stats.uswap   = (long)(k * (m * node->stats.uswap   + qdata->lmd.uswap));
      node->stats.fswap   = (long)(k * (m * node->stats.fswap   + qdata->lmd.fswap));
      node->stats.ts      = qdata->lmd.ts;
      node->stats.trend	  = qdata->lmd.trend;
      node->rb_pos++;
      node->ticks ++;
      
      return;
    }
  }
}


void dump_xpoolmon_data (load_stats **lm_stats_base, procs_hosts_stats *host)
{
#ifdef DO_DUMPS
  int j;
  int k;

  FILE *f;
  int FilenameSize;
  char *Filename;

  /* Open file for writing */
  FilenameSize = strlen(global_options.DumpDir) + 1 + strlen(XDUMP) + 1;
  Filename = (char*)malloc( FilenameSize );
  sprintf( Filename, "%s/%s", global_options.DumpDir, XDUMP );
  if ((f = fopen (Filename, "w")) == NULL) {
    bailout ("dump_xpoolmon_data(), open() failed");
  }
  free( Filename );

  for (j = 0; j < rd_entries; j++) {
    for (k = 0; k < rd_entries; k++) {
      if ((lm_stats_base[j] != NULL) && (host[k].size != 0)) {
	if (lm_stats_base[j]->stats.hid == host[k].stats.hid) {

	  /* Dump data to file */

	  fprintf (f, 
		   "%s %d %f %f %f %f %f %f %d %d %d %d %d\n",
		   lm_stats_base[j]->stats.hname,
		   lm_stats_base[j]->stats.ncpu,
		   lm_stats_base[j]->stats.cpuload,
                   0, /* lm_stats_base[j]->stats.kernel, */
		   0, /* lm_stats_base[j]->stats.wait, */
		   0, /* lm_stats_base[j]->stats.user, */
		   host[k].stats.pct_cpu,
		   0, /* lm_stats_base[j]->stats.idle, */
		   lm_stats_base[j]->stats.uram,
		   host[k].stats.mem_img,
		   lm_stats_base[j]->stats.fram,
		   lm_stats_base[j]->stats.uswap,
		   lm_stats_base[j]->stats.fswap
		   );
	}
      }
    }
  }
  fprintf (f, "[EOF]\n");
  fclose (f);
#endif
}


void dump_LOAD_stats (load_stats **lm_stats_base)
{
#ifdef DO_DUMPS
  int i;
  int j;
  FILE *f;
  char* Filename;
  int FilenameSize;
     
  /* Open file for writing */
  FilenameSize = strlen(global_options.DumpDir) + 1 + strlen(LDUMP) + 1;
  Filename = (char*)malloc( FilenameSize );
  sprintf( Filename, "%s/%s", global_options.DumpDir, LDUMP );
  if ((f = fopen (Filename, "w")) == NULL) {
    bailout ("dump_load_stat(), open() failed");
  }
  free( Filename );


  /* Dump data to file */
  fprintf (f, "#Host  HID  #CPU   load     kernel     wait      user      idle    uram   fram   uswap  fswap    ts\n");
  for (i = 0; 
       (i < rd_entries) && (lm_stats_base[i] != NULL); 
       i++) {
    fprintf (f, 
	     "%.6s %3d   %3d %9.6f %9.6f %9.6f %9.6f %9.6f %6d %6d %6d %6d %6d\n",
	     lm_stats_base[i]->stats.hname,
	     lm_stats_base[i]->stats.hid,
	     lm_stats_base[i]->stats.ncpu,
	     lm_stats_base[i]->stats.cpuload,
	     0, /* lm_stats_base[i]->stats.kernel, */
	     0, /* lm_stats_base[i]->stats.wait, */
	     0, /* lm_stats_base[i]->stats.user, */
	     0, /* lm_stats_base[i]->stats.idle, */
	     lm_stats_base[i]->stats.uram,
	     lm_stats_base[i]->stats.fram,
	     lm_stats_base[i]->stats.uswap,
	     lm_stats_base[i]->stats.fswap,
	     lm_stats_base[i]->stats.ts
	     );
  }
  fprintf (f, "\n");

  /* Ring buffer details */
  for (i = 0; 
       (i < rd_entries) && (lm_stats_base[i] != NULL); 
       i++) {
    fprintf (f, 
	     "\n#**** Ring buffer details ****\n");
    fprintf (f, "#Host  HID  #CPU   load     kernel     wait      user      idle    uram   fram   uswap  fswap    ts\n");
    for (j = 0; j < lm_stats_base[i]->ticks; j++) {
      fprintf (f, 
	       "%.6s %3d   %3d %9.6f %9.6f %9.6f %9.6f %9.6f %6d %6d %6d %6d %6d\n",
	       lm_stats_base[i]->rb_head[j].hname,
	       lm_stats_base[i]->rb_head[j].hid,
	       lm_stats_base[i]->rb_head[j].ncpu,
	       lm_stats_base[i]->rb_head[j].cpuload,
	       0, /* lm_stats_base[i]->rb_head[j].kernel, */
	       0, /* lm_stats_base[i]->rb_head[j].wait, */
	       0, /* lm_stats_base[i]->rb_head[j].user, */
	       0, /* lm_stats_base[i]->rb_head[j].idle, */
	       lm_stats_base[i]->rb_head[j].uram,
	       lm_stats_base[i]->rb_head[j].fram,
	       lm_stats_base[i]->rb_head[j].uswap,
	       lm_stats_base[i]->rb_head[j].fswap,
	       lm_stats_base[i]->rb_head[j].ts
	       );
    }
  }
  fprintf (f, "[EOF]\n");

  fclose (f);
#endif
}


void dump_PROCS_stats (procs_hosts_stats *host)
{
#ifdef DO_DUMPS
  int i;
  int j;
  int k;
  FILE *f;
  char* Filename;
  int FilenameSize;

  /* Open file for writing */
  FilenameSize = strlen(global_options.DumpDir) + 1 + strlen(PDUMP) + 1;
  Filename = (char*)malloc( FilenameSize );
  sprintf( Filename, "%s/%s", global_options.DumpDir, PDUMP );
  if ((f = fopen (Filename, "w")) == NULL) {
    bailout ("dump_PROCS_stat(), open() failed");
  }
  free( Filename );

  /* Dump data to file */
  for (i = 0; 
       (i < rd_entries) && (host[i].size != 0); 
       i++) {
    fprintf (f, 
	     "\nHost: %s    Accumulated average PVM CPU: %f%%    PVM Memory: %d MB\n",
	     host[i].stats.hname,
	     host[i].stats.pct_cpu * 100,
	     host[i].stats.mem_img / 1024);
    for (j = 0; j < host[i].size; j++) {
      if( !host[i].base[j].task_is_dead ) {
        fprintf (f,
  	       "\tHID: %3d  TID: %6d  PID: %6d  PVM Avrg CPU: %7.4f%%  Avrg IMG: %6d MB\n",
  	       host[i].base[j].stats.hid,
  	       host[i].base[j].stats.tid,
  	       host[i].base[j].stats.pid,
  	       host[i].base[j].stats.pct_cpu * 100,
  	       host[i].base[j].stats.mem_img / 1024);
      }
    }
  }
  fprintf (f, "\n");
  
  /* Ring buffer details */
  for (i = 0; 
       (i < rd_entries) && (host[i].size != 0); 
       i++) {
    for (j = 0; j < host[i].size; j++) {
      if( !host[i].base[j].task_is_dead ) {
        fprintf (f, "\n**** Ring buffer details ****\n");
        fprintf (f, "Host  HID    TID     PID       CPU       IMG      ts\n");
        for (k = 0; k < host[i].base[j].ticks; k++) {
          if( !host[i].base[j].rb_head[k].hname )       continue;
  	  fprintf (f, 
		 "%.5s %3d   %6d  %5d   %9.6f   %6d   %d\n",
		 host[i].base[j].rb_head[k].hname,
		 host[i].base[j].rb_head[k].hid,
		 host[i].base[j].rb_head[k].tid,
		 host[i].base[j].rb_head[k].pid,
		 host[i].base[j].rb_head[k].pct_cpu,
		 host[i].base[j].rb_head[k].mem_img,
		 host[i].base[j].rb_head[k].ts
		 );
        }
      }
    }
  }
  fprintf (f, "[EOF]\n");

  fclose (f);
#endif
}


int copy_PROCS_data (procs_data *to, procs_data *from)
{
  /* Copy the data */
  memcpy (to, from, sizeof (procs_data));
  if ((to->hname = strdup (from->hname)) == NULL) {
    fprintf (stderr, "copy_PROCS_data(), strdup() failed\n");
    return (-1);
  }
  
  return (0);
}


/* Currently implemented is the mean average of a time window */
void update_PROCS_stats (procs_hosts_stats *host, 
			 int i, 
			 int j,
			 ll_procs_queue *qdata)
{
  double d;
  double k;
  int    l;
  int    m;
#ifdef _DEBUG
  cerr<<"i: "<<i<<", j: "<<j<<" --> size: "<<host[i].size<<endl;
#endif
  /* New statistics for host i? */
  if (host[i].size == 0) {

    /* Yes, simply copy data for host statistics */
    if (copy_PROCS_data (&(host[i].stats), &(qdata->pmd)) < 0) {
      bailout ("update_PROCS_stats(), copy_PROCS_data failed (1)");
    }

    /* Clear useless entries */
    host[i].stats.tid = 0;
    host[i].stats.pid = 0;
    host[i].stats.ts  = 0;

    /* Copy data for process statistics */
    if (copy_PROCS_data (&(host[i].base[j].stats), &(qdata->pmd)) < 0) {
      bailout ("update_PROCS_stats(), copy_PROCS_data failed (2)");
    }

    /* Preserve a copy of the data */
    if (copy_PROCS_data (host[i].base[j].rb_head, &(qdata->pmd)) < 0) {
      bailout ("update_PROCS_stats(), copy_PROCS_data failed (3)");
    }

    /* Update ring buffer pointer */
    host[i].base[j].rb_pos++;

    /* Update ticks for process */
    host[i].base[j].ticks++;

    host[i].base[j].stats.movable = qdata->pmd.movable;

  } else {

    /* No, check for ticks of timewindow are complete */
    if (host[i].base[j].ticks  == DEFMONWT) {

      if (host[i].base[j].rb_pos > host[i].base[j].rb_tail) {
	/* We would write behind the array, correct this */
	host[i].base[j].rb_pos = host[i].base[j].rb_head;
      }

      /* _           _         1
       * x(t=t   ) = x(t=t ) - - ( x(t=t - nT) - x(t=t   ))
       *      j+1         j    n        j             j+1
       */

      k = 1.0 / (double) DEFMONWT;    /* 1/n */
    
      /* Statistics of the process */
      host[i].base[j].stats.movable = qdata->pmd.movable;

      host[i].base[j].stats.pct_cpu -= 
	k * (host[i].base[j].rb_pos->pct_cpu - qdata->pmd.pct_cpu);

      host[i].base[j].stats.pct_mem -= 
	k * (host[i].base[j].rb_pos->pct_mem - qdata->pmd.pct_mem);

      /* Attention unsigned long ahed, negativ values are not good ... */
      d = (double) host[i].base[j].stats.mem_img;
      d -= k * (((double) host[i].base[j].rb_pos->mem_img) - 
		((double) qdata->pmd.mem_img));
      host[i].base[j].stats.mem_img = (unsigned long) d;
      d = (double) host[i].base[j].stats.mem_res;
      d -= k * (((double) host[i].base[j].rb_pos->mem_res) - 
		((double) qdata->pmd.mem_res));
      host[i].base[j].stats.mem_res = (unsigned long) d;
      host[i].base[j].stats.ts  = qdata->pmd.ts;
            
      /* update the total_pct_cpu and totals variables */
      host[i].base[j].stats.total_pct_cpu=
    		((host[i].base[j].stats.total_pct_cpu*host[i].base[j].stats.totals)+
		 (qdata->pmd.pct_cpu*qdata->pmd.totals))/
		(host[i].base[j].stats.totals+qdata->pmd.totals);
      host[i].base[j].stats.totals+=qdata->pmd.totals;
    
      /* Statistics of the host */
      host[i].stats.pct_cpu = 0;
      host[i].stats.pct_mem = 0;
      host[i].stats.mem_img = 0;
      host[i].stats.mem_res = 0;
      for (l = 0; l < host[i].size; l++) {
        if( !host[i].base[l].task_is_dead ) {
    	  host[i].stats.pct_cpu += host[i].base[l].stats.pct_cpu;
	  host[i].stats.pct_mem += host[i].base[l].stats.pct_mem;
  	  host[i].stats.mem_img += host[i].base[l].stats.mem_img;
	  host[i].stats.mem_res += host[i].base[l].stats.mem_res;
        }
      }

      /* Free the dynamic data of the current entry, its obsolete */
      free (host[i].base[j].rb_pos->hname);

      /* Preserve a copy of the data */
      if (copy_PROCS_data (host[i].base[j].rb_pos, &(qdata->pmd)) < 0) {
	bailout ("update_PROCS_stats(), copy_PROCS_data failed (4)");
      }

      /* Update ring buffer pointer */
      host[i].base[j].rb_pos++;
      
    } else {

      /* Check for new process on host i */
      if (host[i].base[j].ticks == 0) {

	/* Yes, simply copy data for process statistics */
	if (copy_PROCS_data (&(host[i].base[j].stats), &(qdata->pmd)) < 0) {
	  bailout ("update_PROCS_stats(), copy_PROCS_data failed (5)");
	}

	/* Preserve a copy of the data */
	if (copy_PROCS_data (host[i].base[j].rb_head, &(qdata->pmd)) < 0) {
	  bailout ("update_PROCS_stats(), copy_PROCS_data failed (6)");
	}

	/* Update ring buffer pointer */
	host[i].base[j].rb_pos++;

	/* Update ticks for process */
	host[i].base[j].ticks++;
        host[i].base[j].stats.movable = qdata->pmd.movable;

      } else {
      
	/* Preserve a copy of the data */
	if (copy_PROCS_data (host[i].base[j].rb_pos, &(qdata->pmd)) < 0) {
	  bailout ("update_PROCS_stats(), copy_PROCS_data failed (7)");
	}

	/* _         1         _
	 * x(t=t ) = - ( (j-1) x(t=t   ) + x(t=t ) )
	 *      j    j              j-1         j
	 */

	k = 1.0 / (double) (host[i].base[j].ticks + 1);  /* 1/j */
	m = host[i].base[j].ticks;                       /* j-1 */
        host[i].base[j].stats.movable = qdata->pmd.movable;

	host[i].base[j].stats.pct_cpu = 
	  k * (m * host[i].base[j].stats.pct_cpu + qdata->pmd.pct_cpu);

	host[i].base[j].stats.pct_mem = 
	  k * (m * host[i].base[j].stats.pct_mem + qdata->pmd.pct_mem);
	host[i].base[j].stats.mem_img = 
	  (long)(k * (m * host[i].base[j].stats.mem_img + qdata->pmd.mem_img));
	host[i].base[j].stats.mem_res = 
	  (long)(k * (m * host[i].base[j].stats.mem_res + qdata->pmd.mem_res));
	host[i].base[j].stats.ts  = qdata->pmd.ts;


        /* update the total_pct_cpu and totals variables */
        host[i].base[j].stats.total_pct_cpu=
    		((host[i].base[j].stats.total_pct_cpu*host[i].base[j].stats.totals)+
		 (qdata->pmd.pct_cpu*qdata->pmd.totals))/
		(host[i].base[j].stats.totals+qdata->pmd.totals);
        host[i].base[j].stats.totals+=qdata->pmd.totals;

	/* Statistics of the host, */
	host[i].stats.pct_cpu = 0;
	host[i].stats.pct_mem = 0;
	host[i].stats.mem_img = 0;
	host[i].stats.mem_res = 0;
	for (l = 0; l < host[i].size; l++) {
          if( !host[i].base[l].task_is_dead ) {
	    host[i].stats.pct_cpu += host[i].base[l].stats.pct_cpu;
	    host[i].stats.pct_mem += host[i].base[l].stats.pct_mem;
	    host[i].stats.mem_img += host[i].base[l].stats.mem_img;
	    host[i].stats.mem_res += host[i].base[l].stats.mem_res;
          }
	}

	/* Update ring buffer pointer */
	host[i].base[j].rb_pos++;
	
	/* Update ticks for process */
	host[i].base[j].ticks++;
        host[i].base[j].stats.movable = qdata->pmd.movable;
      }
    }
  }
  
  /* now have a loook if there wasn't a task with this tid in the "wait for dead" queue ... */
  bal_queue_search(host[i].base[j].stats.tid,&(host[i].base[j].stats));
}


void update_PROCS_termination (procs_hosts_stats *host, 
			 int i, 
			 int j,
			 ll_procs_queue *qdata)
{
  double d;
  double k;
  int    l;
  int    m;
#ifdef _DEBUG
  cerr<<"i: "<<i<<", j: "<<j<<" --> size: "<<host[i].size<<endl;
#endif

  
  /* this is a termination event, enter this task into the "wait for dead" queue ... */

#ifdef _DEBUG
  cerr<<"termination event i: "<<i<<" j: "<<j<<" pname: "<<host[i].base[j].stats.pname;
#endif
  bal_queue_insert(host[i].base[j].stats.pname,host[i].base[j].stats.tid,host[i].base[j].stats.total_pct_cpu,host[i].base[j].stats.totals,host[i].base[j].stats.mem_res);
}

void *process_COMM_queue (void *p)
{
  ThreadIgnoreSigAlarm();

  /* Temporary pointer for head update operation */
  ll_comm_queue *tmp_ll_cq; 

  FILE *f;
  int FilenameSize = strlen(global_options.DumpDir) + 1 + strlen(XDUMP) + 1;
  char*  Filename = (char*)malloc( FilenameSize );
  sprintf( Filename, "%s/%s", global_options.DumpDir, CDUMP );

  /* We will never return */
  while (1) {

    /* Try to aquire the mutex for the communication queue */
    pthread_mutex_lock (&cq_mutex);

    /* Wait until there is data in the communication queue */
    pthread_cond_wait (&cq_cond, &cq_mutex);

    /* Append to file */
    if ((f = fopen (Filename, "w")) == NULL) {
      bailout ("process_COMM_queue(), open() failed");
    }
    
    /* Process all entries in the communication queue */
    while (commq.head != NULL) {
 
      /*
       * Do something useful with the data.
       */
      
      /* Dump data to file */
      fprintf (f, 
	       "%s: %x -> %x type %d len %d start %ld.%ld end %ld.%ld\n",
	       montypes[commq.head->cmd.mon_montype],
	       commq.head->cmd.mon_src, 
	       commq.head->cmd.mon_dst, 
	       commq.head->cmd.mon_msgtype,
	       commq.head->cmd.mon_len, 
	       commq.head->cmd.mon_stime.tv_sec,
	       commq.head->cmd.mon_stime.tv_nsec, 
	       commq.head->cmd.mon_etime.tv_sec,
	       commq.head->cmd.mon_etime.tv_nsec);
    
      /* Save the pointer to the next entry */
      tmp_ll_cq = commq.head->next;
      
      /* Free the dynamically allocated members of 
	 the communication message data structure */
      /* No special action required in this case */

      /* Free the first entry from the list */
      free ((void *) commq.head);
      
      /* Was it the only one */
      if (tmp_ll_cq == NULL) {
	
	/* Now we have an empty communication queue */
	commq.head = NULL;
	commq.tail = NULL;
	
      } else {
	
	/* There is data left, update pointer to head of communication queue */
	commq.head = tmp_ll_cq;
      }
    }

    /* Close the file */
    fprintf (f, "[EOF]\n");
    fclose (f);

    /* Free the mutex for the communication queue */
    pthread_mutex_unlock (&cq_mutex);

    /* Wait a little time before next run */
    sleep (DEFCOMMWAIT);
  }
  return 0;
}


void *process_LOAD_queue (void *P)
{
  ThreadIgnoreSigAlarm();

  /* Temporary pointer for head update operation */
  ll_load_queue *tmp_ll_lq; 


  /* Flag to signal entry found */
  int found;
  
  int i;

  /* Allocate memory for the pointers to the load statistics of each host */
  if ((lm_stats_base = 
       (load_stats **) calloc (rd_entries, sizeof (load_stats *))) == NULL) {
    bailout ("process_LOAD_queue(), calloc() for lm_stat_base failed");
  }

  /* We will never return */
  while (1) {

    /* Try to aquire the mutex for the load queue */
    pthread_mutex_lock (&lq_mutex);

    /* Wait until there is data in the load queue */
    pthread_cond_wait (&lq_cond, &lq_mutex);

    /* Try to aquire the mutex for the load statistics */
    pthread_mutex_lock (&ls_mutex);
    
    /* Process all entries in the load queue */
    while (loadq.head != NULL) {
    
      /*
       * Do something useful with the data.
       */

      /* Which host send us the data? */
      for (i = 0, found = 0; 
	   (i < rd_entries) && (lm_stats_base[i] != NULL); 
	   i++) {
	if (lm_stats_base[i]->stats.hid == loadq.head->lmd.hid) {
	  found = 1;
	  break;
	}
      }
      
      /* Host already known? */
      if (found) {
	
	/* Yes, calc average for host */
	update_LOAD_stats (lm_stats_base[i], loadq.head); 

      } else {

	/* No, check for index out of bounds */
	if (i < rd_entries) {

	  /* Everything is fine, allocate memory for new host*/
	  if ((lm_stats_base[i] = 
	       (load_stats *) calloc (1, sizeof (load_stats))) == NULL) {
	    bailout ("process_LOAD_queue(), calloc() for lm_stats_base[i] failed");
	  }     
	  
	  /* Calculate average for new host */
	  update_LOAD_stats (lm_stats_base[i], loadq.head);

	} else {
	  
	  /* Index out of range */
	  bailout ("process_LOAD_queue(), index out of range");
	}
      }
      
      /* Save the pointer to the next entry */
      tmp_ll_lq = loadq.head->next;
      
      /* Free the members of the load message data structure */
      free ((void *) loadq.head->lmd.hname);
      
      /* Free the first entry from the list */
      free ((void *) loadq.head);
    
      /* Was it the only one */
      if (tmp_ll_lq == NULL) {
	
	/* Now we have an empty load queue */
	loadq.head = NULL;
	loadq.tail = NULL;
      
      } else {
	
	/* There is data left, update pointer to head of load queue */
	loadq.head = tmp_ll_lq;
      }
    }
    
    /* Dump statistics to file */
    dump_LOAD_stats (lm_stats_base);

    /* Free the mutex for the load statistics */
    pthread_mutex_unlock (&ls_mutex);
    
    /* Free the mutex for the load queue */
    pthread_mutex_unlock (&lq_mutex);

    /* Wait a little time before running again */
    sleep (DEFLOADWAIT);
  }
  return 0;
}


void *process_PROCS_queue (void *P)
{
  ThreadIgnoreSigAlarm();

  /* Temporary pointer for head update operation */
  ll_procs_queue *tmp_ll_pq; 


  /* Flag to signal host entry found */
  int hfound;
  
  /* Flag to signal pid entry found */
  int pfound;
  
  int i;
  int j;

  /* flag to signal, that data is actually a termination notify */
  int termination;

  /* Allocate memory for the pointers 
     to the PROCS statistics of each host */
  if ((pd_hosts_stats = 
       (procs_hosts_stats *) calloc (rd_entries, 
				     sizeof (procs_hosts_stats))) == NULL) {
    bailout ("process_PROCS_queue(), calloc() for pd_hosts_stats failed");
  }

  /* Initialize the host statistics structures */
  for (i = 0; i < rd_entries; i++) {
    pd_hosts_stats[i].max = DEFPROCSALLOC;
  }
  
  /* Allocate default array for PROCS statistics 
     of each process on each host */
  for (i = 0; i < rd_entries; i++) {
    if ((pd_hosts_stats[i].base = (procs_stats *) calloc (DEFPROCSALLOC, sizeof (procs_stats))) == NULL) {
      bailout ("process_PROCS_queue(), calloc() for pd_hosts_stats[i] failed");
    }
  }
  
  /* init the HASH table for task statistics */
  bal_table_init();
  
  /* We will never return */
  while (1) {
#ifdef _DEBUG
    cerr<<getpid()<<"_procs-queue\n";
#endif
    /* Try to aquire the mutex for the procs queue */
    pthread_mutex_lock (&pq_mutex);

    /* Wait until there is data in the procs queue */
    pthread_cond_wait (&pq_cond, &pq_mutex);

    /* Try to aquire the mutex for the process statisitics */
    pthread_mutex_lock (&ps_mutex);

    /* Process all entries in the procs queue */
    while (procsq.head != NULL) {

      /*
       * Do something useful with the data.
       */
      
      /* Which host send us the data? */
      for (i = 0, j = 0, hfound = 0; 
	   (i < rd_entries) && (pd_hosts_stats[i].size > 0); 
	   i++) {
	if (pd_hosts_stats[i].base[0].stats.hid == procsq.head->pmd.hid) {
	  hfound = 1;
	  break;
	}
      }
#ifdef _DEBUG
      cerr<<endl;
      cerr<<procsq.head->pmd.hname<<"-"
	  <<procsq.head->pmd.pid<<"-"
	  <<procsq.head->pmd.pct_cpu<<"-"
	  <<hfound<<flush;
#endif      

      /* is this a termination notify ? */
      if(procsq.head->pmd.movable==-2)termination=1;
        else termination=0;

      /* Host already known? */
      if (hfound) {

	/* Which process on the host send the data? */
	for (j = 0, pfound = 0;
	     j < pd_hosts_stats[i].size;
	     j++) {
	  if (pd_hosts_stats[i].base[j].stats.pid == procsq.head->pmd.pid) {
	    pfound = 1;
	    break;
	  }
	}

	/* Process already known? */
	if (pfound) {
#ifdef _DEBUG
	  cerr<<"-"<<pfound<<flush;
#endif	  
#ifdef _DEBUG
          cerr << "Old task detected: " << procsq.head->pmd.tid << endl;
#endif

        /* Yes, calc average for host and process */
	  if(!termination)update_PROCS_stats (pd_hosts_stats, i, j, procsq.head);
	    else update_PROCS_termination(pd_hosts_stats, i, j, procsq.head);
	  
	} else {
	  
#ifdef _DEBUG
          cerr << "New task detected: " << procsq.head->pmd.tid 
               << " movable==" << procsq.head->pmd.movable << endl;
#endif

	  /* A new process on host i */

	  /* Are there enough entries left to add this process? */
	  if (pd_hosts_stats[i].size < pd_hosts_stats[i].max) {

	    /* Enough space left */

	    /* Allocate memory for the ring buffer of the process */
	    if ((pd_hosts_stats[i].base[j].rb_head = (procs_data *) 
		 calloc (DEFMONWT, sizeof (procs_data))) == NULL) {
	      bailout ("process_PROCS_queue(), calloc for rb_head failed");
	    }
	  
	    /* Initialize ring buffer pointer */
	    pd_hosts_stats[i].base[j].rb_pos = 
	      pd_hosts_stats[i].base[j].rb_head;
	    pd_hosts_stats[i].base[j].rb_tail = 
	      pd_hosts_stats[i].base[j].rb_head + DEFMONWT - 1;

	    /* Calc average for host and process */
	    update_PROCS_stats (pd_hosts_stats, i, j, procsq.head);
	    
	    /* One entry more in array */
	    pd_hosts_stats[i].size++;

	  } else {

	    /* Allocate more memory */
	    if ((pd_hosts_stats[i].base = (procs_stats *) 
		 realloc (pd_hosts_stats[i].base, 
			  (pd_hosts_stats[i].max + DEFPROCSALLOC) * 
			  sizeof (procs_stats))) == NULL) {
	      bailout ("process_PROCS_queue(), realloc() failed");
	    }

	    /* Clear the new part of the array */
	    bzero (pd_hosts_stats[i].base + pd_hosts_stats[i].max, 
		   DEFPROCSALLOC * sizeof (procs_stats));

	    /* Update the array size */
	    pd_hosts_stats[i].max += DEFPROCSALLOC;
	    
	    /* Allocate memory for the ring buffer of the process */
	    if ((pd_hosts_stats[i].base[j].rb_head = (procs_data *) 
		 calloc (DEFMONWT, sizeof (procs_data))) == NULL) {
	      bailout ("process_PROCS_queue(), calloc for rb_head failed");
	    }
	  
	    /* Initialize ring buffer pointer */
	    pd_hosts_stats[i].base[j].rb_pos = 
	      pd_hosts_stats[i].base[j].rb_head;
	    pd_hosts_stats[i].base[j].rb_tail = 
	      pd_hosts_stats[i].base[j].rb_head + DEFMONWT - 1;

	    /* Calc average for host and process */
	    update_PROCS_stats (pd_hosts_stats, i, j, procsq.head);

	    /* One entry more in array */
	    pd_hosts_stats[i].size++;
	  }
	}

      } else {
	
	/* A new host and therefore also a new process */
	printf( "New host detected: %s\n", procsq.head->pmd.hname );

	/* Check for index out of range */
	if (i < rd_entries) {

	  /* Allocate memory for the ring buffer of the process */
	  if ((pd_hosts_stats[i].base[j].rb_head = (procs_data *) 
	       calloc (DEFMONWT, sizeof (procs_data))) == NULL) {
	    bailout ("process_PROCS_queue(), calloc for rb_head failed");
	  }
	  
	  /* Initialize ring buffer pointer */
	  pd_hosts_stats[i].base[j].rb_pos = 
	    pd_hosts_stats[i].base[j].rb_head;
	  pd_hosts_stats[i].base[j].rb_tail = 
	    pd_hosts_stats[i].base[j].rb_head + DEFMONWT - 1;

	  /* Calc average for host and process */
	  update_PROCS_stats (pd_hosts_stats, i, j, procsq.head);
	  
	  /* One entry more in array */
	  pd_hosts_stats[i].size++;
	  
	} else {
	  
	  /* Index out of range */
	  bailout ("process_PROCS_queue(), index out of range");
	}
      }
       
      /* Save the pointer to the next entry */
      tmp_ll_pq = procsq.head->next;
      
      /* Free the members of the procs message data structure */
      free ((void *) procsq.head->pmd.hname);

      /* Free the first entry from the list */
      free ((void *) procsq.head);
    
      /* Was it the only one */
      if (tmp_ll_pq == NULL) {
	
	/* Now we have an empty procs queue */
	procsq.head = NULL;
	procsq.tail = NULL;
      
      } else {
	
	/* There is data left, update pointer to head of procs queue */
	procsq.head = tmp_ll_pq;
      }
    }


    /* write out all the expired "wait for dead tasks" to the HASH table ... */
    bal_queue_purge();
    
    /* Try to aquire the mutex for the load statistics */
    pthread_mutex_lock (&ls_mutex);

    /* Dump file for xpoolmon */
    dump_xpoolmon_data (lm_stats_base, pd_hosts_stats);
    
    /* Free the mutex for the load statistics */
    pthread_mutex_unlock (&ls_mutex);

    /* Dump statistics to file */
    dump_PROCS_stats (pd_hosts_stats);

    /* Free the mutex for the process statistics */
    pthread_mutex_unlock (&ps_mutex);

    /* Free the mutex for the procs queue */
    pthread_mutex_unlock (&pq_mutex);

    /* Wait a little time before running again */
    sleep (DEFPROCSWAIT);
  }
  return 0;
}


/* Helper functions to initialize, duplicate and free a migration_stats 
   structure. ATTENTION this array starts by 1 !!! Range [1...N] */

migration_stats *new_migration_stats(const int amount)
{
  migration_stats *to_ptr;

  if (amount < 1) {
    return (NULL);
  }

  if ((to_ptr = (migration_stats *) 
       calloc (amount + 1, sizeof (migration_stats))) == NULL) {
    bailout ("new_migration_stats(), calloc for to_ptr failed");
  }

  return (to_ptr);
}

migration_stats *dup_migration_stats (const migration_stats *from_ptr, 
				      const int amount) 
{
  migration_stats *to_ptr;
  int i;
  
  if (amount < 1) {
    return (NULL);
  }

  to_ptr = new_migration_stats (amount);

  /* Copy members of the structures */
  for (i = 0; i < amount; i++) {
    *(to_ptr + 1 + i) = *(from_ptr + 1 + i);
  
    /* Copy the dynamic members of each structure */
    if ((from_ptr + 1 + i)->from_hname != NULL) {
      if (((to_ptr + 1 + i)->from_hname = 
	   strdup ((from_ptr + 1 + i)->from_hname)) == NULL) {
	bailout ("dup_migration_stats(), strdup for from_hname failed");
      }
    }
    if ((from_ptr + 1 + i)->to_hname != NULL) {
      if (((to_ptr + 1 + i)->to_hname = 
	   strdup ((from_ptr + 1 + i)->to_hname)) == NULL) {
	bailout ("dup_migration_stats(), strdup for to_hname failed");
      }
    }
  }
  return (to_ptr);
}

void copy_migration_data (migration_stats *to_ptr, 
			  const migration_stats *from_ptr)
{
  /* Free the dynamic members */
  if (to_ptr->from_hname != NULL) {
    free (to_ptr->from_hname);
  }
  if (to_ptr->to_hname != NULL) {
    free (to_ptr->to_hname);
  }

  /* Copy the data */
  *to_ptr = *from_ptr;

  /* Copy the dynamic members */
  if (from_ptr->from_hname != NULL) {
    if ((to_ptr->from_hname = strdup (from_ptr->from_hname)) == NULL) {
      bailout ("copy_migration_data(), strdup for from_hname failed");
    }
  }
  if (from_ptr->to_hname != NULL) {
    if ((to_ptr->to_hname = strdup (from_ptr->to_hname)) == NULL) {
      bailout ("copy_migration_data(), strdup for to_hname failed");
    }
  }
}


void free_migration_stats (migration_stats *from_ptr, const int amount)
{
  int i;
  
  if (amount < 1) {
    return;
  }

  for (i = 0; i < amount; i++) {
    /* Free the dynamic members, then the whole array */
    if ((from_ptr + 1 + i)->from_hname != NULL) {
      free ((from_ptr + 1 + i)->from_hname);
    }
    if ((from_ptr + 1 + i)->to_hname != NULL) {
    free ((from_ptr + 1 + i)->to_hname);
    }
  }
  free (from_ptr);
}



void swap_migration_data (migration_stats *this_ptr, migration_stats *that_ptr)
{
  migration_stats *tmp = new_migration_stats(1);

  copy_migration_data (&tmp[1], this_ptr);
  copy_migration_data (this_ptr, that_ptr);
  copy_migration_data (that_ptr, &tmp[1]);

  free_migration_stats (tmp, 1);
}



void* pvm_connect_thr (void *p)
{
  ThreadIgnoreSigAlarm();

  while( !ConnectedToPVM ) {
    if( pvm_is_running() ) {
      /* Enroll this task into PVM */
      if ( (tid=pvm_mytid()) < 0) {
        bailout ("pvm_connect_thr(), pvm_mytid() failed");
      }
      ConnectedToPVM = 1;

      if ((dtid = pvm_tidtohost(tid)) < 0) {
          bailout ("mon_procs(), pvm_tidtohost() failed");
      }
      
      signal(SIGURG, SIG_IGN);

    } else {
	sleep( 5 );
    }
  }
  return 0;
}


void
remove_task( int taskID, int hostID)
{

    return; //hallo

  int i,k,l;

  /* Remove old load information */
  for( i=0; i<pd_hosts_stats[hostID].size; i++ ) {
      if( pd_hosts_stats[hostID].base[i].stats.tid == taskID ) {

          /* Old data that has already been marked dead, can be released.
             The current entries will be marked dead. */

          if( pd_hosts_stats[hostID].base[i].task_is_dead ) {
              /* free ring buffer */
              free( pd_hosts_stats[hostID].base[i].rb_head );

              /* eliminate entry in the pd_hosts_stats[hostID].base array */
              if( i+1 < pd_hosts_stats[hostID].size ) {
                  memmove( &pd_hosts_stats[hostID].base[i], &pd_hosts_stats[hostID].base[i+1],
                           pd_hosts_stats[hostID].size-1-i );
              }
              --pd_hosts_stats[hostID].size;
          } else {
              pd_hosts_stats[hostID].base[i].task_is_dead = TRUE;
          }

          /* update statistics of the host */
          pd_hosts_stats[hostID].stats.pct_cpu = 0;
          pd_hosts_stats[hostID].stats.pct_mem = 0;
          pd_hosts_stats[hostID].stats.mem_img = 0;
          pd_hosts_stats[hostID].stats.mem_res = 0;
          for (l = 0; l < pd_hosts_stats[hostID].size; l++) {
              if( !pd_hosts_stats[hostID].base[l].task_is_dead ) {
                  pd_hosts_stats[hostID].stats.pct_cpu += 
                      pd_hosts_stats[hostID].base[l].stats.pct_cpu;
                  pd_hosts_stats[hostID].stats.pct_mem += 
                      pd_hosts_stats[hostID].base[l].stats.pct_mem;
                  pd_hosts_stats[hostID].stats.mem_img += 
                      pd_hosts_stats[hostID].base[l].stats.mem_img;
                  pd_hosts_stats[hostID].stats.mem_res += 
                      pd_hosts_stats[hostID].base[l].stats.mem_res;
              }
          }
      }
  }
}


void
move_task( int taskID, int hostID , int newhostID)
{

  int i,k,l,j;

  /* Remove old load information */
#ifdef _DEBUG  
  cerr<<" moving internally taskID "<<taskID<<" from hostid "<<hostID<<" to newhostid "<<newhostID<<endl;
#endif
  for( i=0; i<pd_hosts_stats[hostID].size; i++ ) {
      if( pd_hosts_stats[hostID].base[i].stats.tid == taskID ) {

#ifdef _DEBUG  
	  cerr<<" found taskID at index "<<i<<endl;
#endif
	  
	  j=pd_hosts_stats[newhostID].size;

	  /* Are there enough entries left to add this process? */
	  if (pd_hosts_stats[newhostID].size < pd_hosts_stats[newhostID].max) {

	    /* Enough space left */

	    /* Allocate memory for the ring buffer of the process */
	    if ((pd_hosts_stats[newhostID].base[j].rb_head = (procs_data *) 
		 calloc (DEFMONWT, sizeof (procs_data))) == NULL) {
	      bailout ("process_PROCS_queue(), calloc for rb_head failed");
	    }
	  
	    /* Initialize ring buffer pointer */
	    pd_hosts_stats[newhostID].base[j].rb_pos = 
	      pd_hosts_stats[newhostID].base[j].rb_head;
	    pd_hosts_stats[newhostID].base[j].rb_tail = 
	      pd_hosts_stats[newhostID].base[j].rb_head + DEFMONWT - 1;

	    /* One entry more in array */
	    pd_hosts_stats[newhostID].size++;

	  } else {

	    /* Allocate more memory */
	    if ((pd_hosts_stats[newhostID].base = (procs_stats *) 
		 realloc (pd_hosts_stats[newhostID].base, 
			  (pd_hosts_stats[newhostID].max + DEFPROCSALLOC) * 
			  sizeof (procs_stats))) == NULL) {
	      bailout ("process_PROCS_queue(), realloc() failed");
	    }

	    /* Clear the new part of the array */
	    bzero (pd_hosts_stats[newhostID].base + pd_hosts_stats[newhostID].max, 
		   DEFPROCSALLOC * sizeof (procs_stats));

	    /* Update the array size */
	    pd_hosts_stats[newhostID].max += DEFPROCSALLOC;
	    
	    /* Allocate memory for the ring buffer of the process */
	    if ((pd_hosts_stats[newhostID].base[j].rb_head = (procs_data *) 
		 calloc (DEFMONWT, sizeof (procs_data))) == NULL) {
	      bailout ("process_PROCS_queue(), calloc for rb_head failed");
	    }
	  
	    /* Initialize ring buffer pointer */
	    pd_hosts_stats[newhostID].base[j].rb_pos = 
	      pd_hosts_stats[newhostID].base[j].rb_head;
	    pd_hosts_stats[newhostID].base[j].rb_tail = 
	      pd_hosts_stats[newhostID].base[j].rb_head + DEFMONWT - 1;

	    /* One entry more in array */
	    pd_hosts_stats[newhostID].size++;
	  }
	  
	  strcpy(pd_hosts_stats[newhostID].base[j].stats.pname,pd_hosts_stats[hostID].base[i].stats.pname);
	  pd_hosts_stats[newhostID].base[j].stats.hid=newhostID;
	  pd_hosts_stats[newhostID].base[j].stats.tid=pd_hosts_stats[hostID].base[i].stats.tid;
	  pd_hosts_stats[newhostID].base[j].stats.pct_cpu=pd_hosts_stats[hostID].base[i].stats.pct_cpu;
	  pd_hosts_stats[newhostID].base[j].stats.pct_mem=pd_hosts_stats[hostID].base[i].stats.pct_mem;
	  pd_hosts_stats[newhostID].base[j].stats.mem_img=pd_hosts_stats[hostID].base[i].stats.mem_img;
	  pd_hosts_stats[newhostID].base[j].stats.mem_res=pd_hosts_stats[hostID].base[i].stats.mem_res;
	  
	  pd_hosts_stats[newhostID].base[j].stats.ts=pd_hosts_stats[hostID].base[i].stats.ts;
	  
	  pd_hosts_stats[newhostID].base[j].stats.totals=pd_hosts_stats[hostID].base[i].stats.totals;
	  
	  pd_hosts_stats[newhostID].base[j].stats.total_pct_cpu=pd_hosts_stats[hostID].base[i].stats.total_pct_cpu;
	  
	  pd_hosts_stats[newhostID].base[j].stats.movable=pd_hosts_stats[hostID].base[i].stats.movable;
	  
          /* Old data that has already been marked dead, can be released.
             The current entries will be marked dead. */

          if( pd_hosts_stats[hostID].base[i].task_is_dead ) {
              /* free ring buffer */
              free( pd_hosts_stats[hostID].base[i].rb_head );

              /* eliminate entry in the pd_hosts_stats[hostID].base array */
              if( i+1 < pd_hosts_stats[hostID].size ) {
                  memmove( &pd_hosts_stats[hostID].base[i], &pd_hosts_stats[hostID].base[i+1],
                           pd_hosts_stats[hostID].size-1-i );
              }
              --pd_hosts_stats[hostID].size;
          } else {
              pd_hosts_stats[hostID].base[i].task_is_dead = TRUE;
          }


          /* update statistics of the source host */
          pd_hosts_stats[hostID].stats.pct_cpu = 0;
          pd_hosts_stats[hostID].stats.pct_mem = 0;
          pd_hosts_stats[hostID].stats.mem_img = 0;
          pd_hosts_stats[hostID].stats.mem_res = 0;
          for (l = 0; l < pd_hosts_stats[hostID].size; l++) {
              if( !pd_hosts_stats[hostID].base[l].task_is_dead ) {
                  pd_hosts_stats[hostID].stats.pct_cpu += 
                      pd_hosts_stats[hostID].base[l].stats.pct_cpu;
                  pd_hosts_stats[hostID].stats.pct_mem += 
                      pd_hosts_stats[hostID].base[l].stats.pct_mem;
                  pd_hosts_stats[hostID].stats.mem_img += 
                      pd_hosts_stats[hostID].base[l].stats.mem_img;
                  pd_hosts_stats[hostID].stats.mem_res += 
                      pd_hosts_stats[hostID].base[l].stats.mem_res;
              }
          }

          /* and update statistics of the destination host */
          pd_hosts_stats[newhostID].stats.pct_cpu = 0;
          pd_hosts_stats[newhostID].stats.pct_mem = 0;
          pd_hosts_stats[newhostID].stats.mem_img = 0;
          pd_hosts_stats[newhostID].stats.mem_res = 0;
          for (l = 0; l < pd_hosts_stats[hostID].size; l++) {
              if( !pd_hosts_stats[newhostID].base[l].task_is_dead ) {
                  pd_hosts_stats[newhostID].stats.pct_cpu += 
                      pd_hosts_stats[newhostID].base[l].stats.pct_cpu;
                  pd_hosts_stats[newhostID].stats.pct_mem += 
                      pd_hosts_stats[newhostID].base[l].stats.pct_mem;
                  pd_hosts_stats[newhostID].stats.mem_img += 
                      pd_hosts_stats[newhostID].base[l].stats.mem_img;
                  pd_hosts_stats[newhostID].stats.mem_res += 
                      pd_hosts_stats[newhostID].base[l].stats.mem_res;
              }
          }
      }
  }
}




void *scheduler (void *p) 
{
  int num_hosts;   /* Current number of PVM machines */

  int host;        /* Current examined host */ 
  int proc;        /* Index of the PVM process that causes the highest load */ 
  double max_load; /* Highest load of an PVM process found so far */ 

  int i;
  int j;
  int k;
  int l;
  int m;
  
  /* Endless scheduling */
  while (1) {

    /* Protect statistics against changes */
    pthread_mutex_lock (&ls_mutex);
    pthread_mutex_lock (&ps_mutex);
    pthread_mutex_lock (&hs_mutex);
    
    /* the migration will be stored here ... */
    vector<TMigration> mig_vec;
    
#ifdef _DEBUG
    cerr<<"size of pd_hosts_stats: "<<pd_hosts_stats->size<<endl;
    for (int ii=0; ii<pd_hosts_stats->size; ii++) {
      cerr<<"host "<<ii<<": "<<pd_hosts_stats[ii].size<<endl;
    }
#endif

    /* try to find some migrations */
    /* try to find each time a migration and execute it, but not more than MAX_MIGRATIONS_IN_ONE_RUN times in a row */
    if(FindMigrations(pd_hosts_stats, lm_stats_base, rd_entries,mig_vec,MAX_MIGRATIONS_IN_ONE_RUN)>0){
    
      cout << "Planning to do " << mig_vec.size() << " migrations\n"<<flush;
        for( int l=0; l<mig_vec.size(); l++ ) {
	  cout << "  task " << hex << mig_vec[l].task_id << " from " << mig_vec[l].currentHostName
	       << " to " << mig_vec[l].newHostName << " (" << mig_vec[l].newHostID
	       << ") " << endl << flush;
	}
					    
      if(ConnectedToPVM){
        for( int i=0; i<mig_vec.size(); i++ ) {
            cout << "Moving task " << hex << mig_vec[i].task_id << ".." << flush;
            char tmp[255]; // pvm_move does not accept const char*
            strcpy( tmp, mig_vec[i].newHostName.c_str() );
#ifdef DEBUG
            cerr<<" removed .."<<flush;
#endif
            pvm_move( mig_vec[i].task_id, tmp );
            cout << "[ok]" << endl << flush;
        }
      }else{
          fprintf( stderr, "Migration not done since pvm unavailable.\n" );
      }
    }

    /* Release statistics protection */
    pthread_mutex_unlock (&hs_mutex);
    pthread_mutex_unlock (&ps_mutex);
    pthread_mutex_unlock (&ls_mutex);
    
    /* Wait a little time before next run */
    /* in this modified version, DEFSHEDWAIT has been shortened from 60 to 20 seconds, because
       the modified balancing algorithm is less time consuming */
    sleep (DEFSCHEDWAIT);
  }
  return 0;
}
