#ifndef MONITOR_H
#define MONITOR_H

#include "mon_util.h"


/* Remote shell command */
#ifndef OSTYPE_linux
#define RSHCMD "rsh"
#else
#define RSHCMD "/usr/bin/rsh"
#endif

/* Default filename for LOAD dump */
#define LDUMP "load_dump.txt"
 
/* Default filename for COMM dump */
#define CDUMP "comm_dump.txt"
 
/* Default filename for PROCS dump */
#define PDUMP "procs_dump.txt"
 
/* Default filename for xpoolmon dump */
#define XDUMP "xpoolmon_dump.txt"
 
/* Default receive buffer size */
#define DEFRBUF 2048

/* Default width of the timewindow in ticks */
#define DEFMONWT 3

/* Default depth of array to store PROCS statistics 
   of each processes on each host */
#define DEFPROCSALLOC 10

/* Default time to wait between calculating LOAD statistics */
#define DEFLOADWAIT 10

/* Default time to wait between calculating COMM statistics */
#define DEFCOMMWAIT 5

/* Default time to wait between calculating PROCS statistics */
#define DEFPROCSWAIT 5

/* Default time to wait between two scheduler runs */
/* this is modified from previous 60 to 20 */
#define DEFSCHEDWAIT 20

/* this is the maximum number of migrations to try consequentially before going to sleep for DEFSCHEDWAIT seconds */
#define MAX_MIGRATIONS_IN_ONE_RUN	4

/* uncomment this to enable the load-balancer to collect and actually use the data recorded in the monitor hash-table */
#define USE_HASH_DATA

/* the treshold for considering a node overloaded, set experimentally  */
#define NODE_OVERLOAD_TRESHOLD		2.0

/* this is the ammount of load which has the destination load to be better that the original source load before migration */
#define NODE_MINIMAL_IMPROVE		0.5

/* Default application threshold level */
#define DEFAPPTHR 0.4

/* Globally known types in this module */
extern char *montypes[];

/* Linked list for load queue */
typedef struct ll_load_queue {
  lm_data lmd;
  struct ll_load_queue *next;
} ll_load_queue;

/* Structure for load queue */
typedef struct load_queue {
  ll_load_queue *head;
  ll_load_queue *tail;
} load_queue;

/* Structure for load statistics */
typedef struct load_stats {
  lm_data *rb_head;  /* Ring buffer head */
  lm_data *rb_tail;  /* Ring buffer tail */
  lm_data *rb_pos;   /* Ring buffer current position */
  lm_data stats;     /* Statistics of the node */    
  int ticks;         /* Ticks monitored so far */
} load_stats;

/* Linked list for communication queue */
typedef struct ll_comm_queue {
  struct mon_data cmd;
  struct ll_comm_queue *next;
} ll_comm_queue;

/* Structure for communication queue */
typedef struct comm_queue {
  ll_comm_queue *head;
  ll_comm_queue *tail;
} comm_queue;



/* Linked list for PROCS queue */
typedef struct ll_procs_queue {
  procs_data pmd;
  struct ll_procs_queue *next;
} ll_procs_queue;

/* Structure for PROCS queue */
typedef struct procs_queue {
  ll_procs_queue *head;
  ll_procs_queue *tail;
} procs_queue;

/* Structure for PROCS statistics */
typedef struct procs_stats {
  procs_data *rb_head;  /* Ring buffer head */
  procs_data *rb_tail;  /* Ring buffer tail */
  procs_data *rb_pos;   /* Ring buffer current position */
  procs_data stats;     /* Statistics of the process */    
  int ticks;            /* Ticks monitored so far */
  int task_is_dead;     /* Task has been migrated but there may still be
                           some old procs_mon messages on the wire */
} procs_stats;

/* Structure for PROCS statistics of one host */
typedef struct procs_hosts_stats {
  procs_stats *base;    /* Base pointer to array of PROCS statistics */
  procs_data stats;     /* Statistics of all processes */
  int max;              /* Current maximum of possible array entries */
  int size;             /* Current amount of entries in the array */ 
} procs_hosts_stats;



/* Structure for migration decisions */
typedef struct migration_stats {
  char *from_hname;
  char *to_hname;
  int from_hid;
  int to_hid;
  int from_tid;
  int from_pid;
  double PROCS_cpu;
  double LOAD_cpu;
  double cpu_ratio;   /* PROCS_cpu / LOAD_cpu */
} migration_stats;


/* Number of successfully read resource entries */
extern int rd_entries;

/* The load queue */
extern load_queue loadq;

/* The communication queue */
extern comm_queue commq;

/* The procs queue */
extern procs_queue procsq;
 
/* Base pointer to array of pointers to load statistics of each host */
extern load_stats **lm_stats_base;

/* Base pointer to process statistics of each host */
extern procs_hosts_stats *pd_hosts_stats;

/* Mutex and condition variable for the load queue */
extern pthread_mutex_t lq_mutex; 
extern pthread_cond_t  lq_cond;  

/* Mutex and condition variable for the communication queue */
extern pthread_mutex_t cq_mutex; 
extern pthread_cond_t  cq_cond;  

/* Mutex and condition variable for the procs queue */
extern pthread_mutex_t pq_mutex; 
extern pthread_cond_t  pq_cond;  

/* Mutex for accessing the LOAD statistics (scheduler) */
extern pthread_mutex_t ls_mutex;

/* Mutex for accessing the PROCS statistics (scheduler) */
extern pthread_mutex_t ps_mutex;

/* Mutex for accessing the HISTORY statistics (scheduler) */
extern pthread_mutex_t hs_mutex;

/* Function prototypes of this module */
extern resdes *new_resdes_struct();

extern int read_resdes(resdes **rd_addr, char* Filename);

extern void startup_slaves(resdes *rd_base, char* ConfigFile);

extern void *process_COMM_queue (void *p);

extern void *process_LOAD_queue (void *p);

extern void *process_PROCS_queue (void *p);

extern void *scheduler (void *p);

extern void *pvm_connect_thr( void *p );

#endif /* MONITOR_H */
